from __future__ import annotations

from math import sqrt
from typing import Any

import torch
from torch import Tensor

from pyqtorch.apply import apply_operator_dm
from pyqtorch.embed import Embedding
from pyqtorch.matrices import DEFAULT_MATRIX_DTYPE, IMAT, XMAT, YMAT, ZMAT
from pyqtorch.utils import (
    DensityMatrix,
    density_mat,
    promote_operator,
    qubit_support_as_tuple,
)


class Noise(torch.nn.Module):
    def __init__(
        self,
        kraus: list[Tensor],
        target: int,
        error_probability: tuple[float, ...] | float,
    ) -> None:
        super().__init__()
        self.target: int = target
        self.qubit_support: tuple[int, ...] = qubit_support_as_tuple(self.target)
        self.is_diagonal = False
        for index, tensor in enumerate(kraus):
            self.register_buffer(f"kraus_{index}", tensor)
        self._device: torch.device = kraus[0].device
        self._dtype: torch.dtype = kraus[0].dtype
        self.error_probability: tuple[float, ...] | float = error_probability

    def extra_repr(self) -> str:
        return f"target: {self.qubit_support}, prob: {self.error_probability}"

    @property
    def kraus_operators(self) -> list[Tensor]:
        return [getattr(self, f"kraus_{i}") for i in range(len(self._buffers))]

    def tensor(self, n_qubit_support: int | None = None) -> list[Tensor]:
        # Since PyQ expects tensor.Size = [2**n_qubits, 2**n_qubits,batch_size].
        t_ops = [kraus_op.unsqueeze(2) for kraus_op in self.kraus_operators]
        if n_qubit_support is None:
            return t_ops
        return [promote_operator(t, self.target, n_qubit_support) for t in t_ops]

    def forward(
        self,
        state: Tensor,
        values: dict[str, Tensor] | Tensor | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        """
        Applies a noisy quantum channel on the input state.
        The evolution is represented as a sum of Kraus operators:
        .. math::
            S(\\rho) = \\sum_i K_i \\rho K_i^\\dagger,

        Each Kraus operator in the `kraus_list` is applied to the input state, and the result
        is accumulated to obtain the evolved state.

        Args:
            state (Tensor): Input quantum state represented as a tensor.

        Returns:
            Tensor: Quantum state as a density matrix after evolution.

        Raises:
            TypeError: If the input `state` or `kraus_list` is not a Tensor.
        """
        values = values or dict()
        if not isinstance(state, DensityMatrix):
            state = density_mat(state)
        rho_evols: list[Tensor] = []
        for kraus in self.tensor():
            rho_evol: Tensor = apply_operator_dm(state, kraus, self.qubit_support)
            rho_evols.append(rho_evol)
        rho_final: Tensor = torch.stack(rho_evols, dim=0)
        rho_final = torch.sum(rho_final, dim=0)
        return rho_final

    @property
    def device(self) -> torch.device:
        return self._device

    @property
    def dtype(self) -> torch.dtype:
        return self._dtype

    def to(self, *args: Any, **kwargs: Any) -> Noise:
        super().to(*args, **kwargs)
        self._device = self.kraus_0.device
        self._dtype = self.kraus_0.dtype
        return self


class BitFlip(Noise):
    """
    Initialize the BitFlip gate.

    The bit flip channel is defined as:

    .. math::
        \\rho \\Rightarrow (1-p) \\rho + p X \\rho X^{\\dagger}

    Args:
        target (int): The index of the qubit being affected by the noise.
        error_probability (float): The probability of a bit flip error.

    Raises:
        TypeError: If the error_probability value is not a float.
    """

    def __init__(
        self,
        target: int,
        error_probability: float,
    ):
        if error_probability > 1.0 or error_probability < 0.0:
            raise ValueError("The error_probability value is not a correct probability")
        K0: Tensor = sqrt(1.0 - error_probability) * IMAT
        K1: Tensor = sqrt(error_probability) * XMAT
        kraus_bitflip: list[Tensor] = [K0, K1]
        super().__init__(kraus_bitflip, target, error_probability)


class PhaseFlip(Noise):
    """
    Initialize the PhaseFlip gate

    The phase flip channel is defined as:

    .. math::
        \\rho \\Rightarrow (1-p) \\rho + p Z \\rho Z^{\\dagger}

    Args:
        target (int): The index of the qubit being affected by the noise.
        error_probability (float): The probability of phase flip error.

    Raises:
        TypeError: If the error_probability value is not a float.
    """

    def __init__(
        self,
        target: int,
        error_probability: float,
    ):
        if error_probability > 1.0 or error_probability < 0.0:
            raise ValueError("The error_probability value is not a correct probability")
        K0: Tensor = sqrt(1.0 - error_probability) * IMAT
        K1: Tensor = sqrt(error_probability) * ZMAT
        kraus_phaseflip: list[Tensor] = [K0, K1]
        super().__init__(kraus_phaseflip, target, error_probability)


class Depolarizing(Noise):
    """
    Initialize the Depolarizing gate.

    The depolarizing channel is defined as:

    .. math::
        \\rho \\Rightarrow (1-p) \\rho
            + p/3 X \\rho X^{\\dagger}
            + p/3 Y \\rho Y^{\\dagger}
            + p/3 Z \\rho Z^{\\dagger}

    Args:
        target (int): The index of the qubit being affected by the noise.
        error_probability (float): The probability of depolarizing error.

    Raises:
        TypeError: If the error_probability value is not a float.
    """

    def __init__(
        self,
        target: int,
        error_probability: float,
    ):
        if error_probability > 1.0 or error_probability < 0.0:
            raise ValueError("The error_probability value is not a correct probability")
        K0: Tensor = sqrt(1.0 - error_probability) * IMAT
        K1: Tensor = sqrt(error_probability / 3) * XMAT
        K2: Tensor = sqrt(error_probability / 3) * YMAT
        K3: Tensor = sqrt(error_probability / 3) * ZMAT
        kraus_depolarizing: list[Tensor] = [K0, K1, K2, K3]
        super().__init__(kraus_depolarizing, target, error_probability)


class PauliChannel(Noise):
    """
    Initialize the PauliChannel gate.

    The pauli channel is defined as:

    .. math::
        \\rho \\Rightarrow (1-px-py-pz) \\rho
            + px X \\rho X^{\\dagger}
            + py Y \\rho Y^{\\dagger}
            + pz Z \\rho Z^{\\dagger}

    Args:
        target (int): The index of the qubit being affected by the noise.
        error_probability (Tuple[float, ...]): Tuple containing probabilities
            of X, Y, and Z errors.

    Raises:
        TypeError: If the probabilities values are not provided in a tuple.
    """

    def __init__(
        self,
        target: int,
        error_probability: tuple[float, ...],
    ):
        sum_prob = sum(error_probability)
        if sum_prob > 1.0:
            raise ValueError("The sum of probabilities can't be greater than 1.0")
        for probability in error_probability:
            if probability > 1.0 or probability < 0.0:
                raise ValueError("The probability values are not correct probabilities")
        px, py, pz = (
            error_probability[0],
            error_probability[1],
            error_probability[2],
        )
        K0: Tensor = sqrt(1.0 - (px + py + pz)) * IMAT
        K1: Tensor = sqrt(px) * XMAT
        K2: Tensor = sqrt(py) * YMAT
        K3: Tensor = sqrt(pz) * ZMAT
        kraus_pauli_channel: list[Tensor] = [K0, K1, K2, K3]
        super().__init__(kraus_pauli_channel, target, error_probability)


class AmplitudeDamping(Noise):
    """
    Initialize the AmplitudeDamping gate.

    The amplitude damping channel is defined as:

    .. math::
        \\rho \\Rightarrow K_0 \\rho K_0^{\\dagger} + K_1 \\rho K_1^{\\dagger}

    with:

    .. code-block:: python

        K0 = [[1, 0], [0, sqrt(1 - rate)]]
        K1 = [[0, sqrt(rate)], [0, 0]]

    Args:
        target (int): The index of the qubit being affected by the noise.
        error_probability (float): The damping rate, indicating the probability of amplitude loss.

    Raises:
        TypeError: If the rate value is not a float.
        ValueError: If the damping rate is not a correct probability.
    """

    def __init__(
        self,
        target: int,
        error_probability: float,
    ):
        rate = error_probability
        if rate > 1.0 or rate < 0.0:
            raise ValueError("The damping rate is not a correct probability")
        K0: Tensor = torch.tensor(
            [[1, 0], [0, sqrt(1 - rate)]], dtype=DEFAULT_MATRIX_DTYPE
        )
        K1: Tensor = torch.tensor([[0, sqrt(rate)], [0, 0]], dtype=DEFAULT_MATRIX_DTYPE)
        kraus_amplitude_damping: list[Tensor] = [K0, K1]
        super().__init__(kraus_amplitude_damping, target, rate)


class PhaseDamping(Noise):
    """
    Initialize the PhaseDamping gate.

    The phase damping channel is defined as:

    .. math::
        \\rho \\Rightarrow K_0 \\rho K_0^{\\dagger} + K_1 \\rho K_1^{\\dagger}

     with:

    .. code-block:: python

        K0 = [[1, 0], [0, sqrt(1 - rate)]]
        K1 = [[0, 0], [0, sqrt(rate)]]

    Args:
        target (int): The index of the qubit being affected by the noise.
        error_probability (float): The damping rate, indicating the probability of phase damping.

    Raises:
        TypeError: If the rate value is not a float.
        ValueError: If the damping rate is not a correct probability.
    """

    def __init__(
        self,
        target: int,
        error_probability: float,
    ):
        rate = error_probability
        if rate > 1.0 or rate < 0.0:
            raise ValueError("The damping rate is not a correct probability")
        K0: Tensor = torch.tensor(
            [[1, 0], [0, sqrt(1 - rate)]], dtype=DEFAULT_MATRIX_DTYPE
        )
        K1: Tensor = torch.tensor([[0, 0], [0, sqrt(rate)]], dtype=DEFAULT_MATRIX_DTYPE)
        kraus_phase_damping: list[Tensor] = [K0, K1]
        super().__init__(kraus_phase_damping, target, rate)


class GeneralizedAmplitudeDamping(Noise):
    """
    Initialize the GeneralizeAmplitudeDamping gate.

    The generalize amplitude damping channel is defined as:

    .. math::
        \\rho \\Rightarrow K_0 \\rho K_0^{\\dagger} + K_1 \\rho K_1^{\\dagger}
            + K_2 \\rho K_2^{\\dagger} + K_3 \\rho K_3^{\\dagger}

    with:

    .. code-block:: python

        K0 = sqrt(p) * [[1, 0], [0, sqrt(1 - rate)]]
        K1 = sqrt(p) * [[0, sqrt(rate)], [0, 0]]
        K2 = sqrt(1-p) * [[sqrt(1 - rate), 0], [0, 1]]
        K3 = sqrt(1-p) * [[0, 0], [sqrt(rate), 0]]

    Args:
        target (int): The index of the qubit being affected by the noise.
        error_probabilities (tuple[float,...]): The first float must be the probability
            of amplitude damping error, and the second float is the damping rate, indicating
            the probability of generalized amplitude damping.

    Raises:
        TypeError: If the probability or rate values is not float.
        ValueError: If the damping rate  a correct probability.
    """

    def __init__(
        self,
        target: int,
        error_probability: tuple[float, ...],
    ):
        probability = error_probability[0]
        rate = error_probability[1]
        if probability > 1.0 or probability < 0.0:
            raise ValueError("The probability value is not a correct probability")
        if rate > 1.0 or rate < 0.0:
            raise ValueError("The damping rate is not a correct probability")
        K0: Tensor = sqrt(probability) * torch.tensor(
            [[1, 0], [0, sqrt(1 - rate)]], dtype=DEFAULT_MATRIX_DTYPE
        )
        K1: Tensor = sqrt(probability) * torch.tensor(
            [[0, sqrt(rate)], [0, 0]], dtype=DEFAULT_MATRIX_DTYPE
        )
        K2: Tensor = sqrt(1.0 - probability) * torch.tensor(
            [[sqrt(1 - rate), 0], [0, 1]], dtype=DEFAULT_MATRIX_DTYPE
        )
        K3: Tensor = sqrt(1.0 - probability) * torch.tensor(
            [[0, 0], [sqrt(rate), 0]], dtype=DEFAULT_MATRIX_DTYPE
        )
        kraus_generalized_amplitude_damping: list[Tensor] = [K0, K1, K2, K3]
        super().__init__(kraus_generalized_amplitude_damping, target, error_probability)



from __future__ import annotations

from abc import ABC
from collections import Counter
from enum import Enum
from functools import singledispatchmethod
from math import log

import torch
from torch import Tensor
from torch.distributions import normal, poisson, uniform


class WhiteNoise(Enum):
    """White noise distributions."""

    UNIFORM = staticmethod(uniform.Uniform(low=0.0, high=1.0))
    """Uniform white noise."""

    GAUSSIAN = staticmethod(normal.Normal(loc=0.0, scale=1.0))
    """Gaussian white noise."""

    POISSON = staticmethod(poisson.Poisson(rate=0.1))
    """Poisson white noise."""


def bitstring_to_tensor(bitstring: str) -> Tensor:
    """
    A helper function to convert bit strings to torch.Tensor.

    Args:
        bitstring:  A str format of a bit string.

    Returns:
        A torch.Tensor out of the input bit string.
    """
    return torch.as_tensor(list(map(int, bitstring)))


def tensor_to_bitstring(bitstring: Tensor) -> str:
    """
    A helper function to convert torch.Tensor to bit strings.

    Args:
        bitstring: A torch.Tensor format of a bit string.

    Returns:
        A str out of the input bit string.
    """
    return "".join(list(map(str, bitstring.detach().tolist())))


def sample_to_matrix(sample: dict) -> Tensor:
    """
    A helper function that maps a sample dict to a bit string array.

    Args:
        sample: A dictionary with bitstrings as keys and values
        as their counts.

    Returns: A torch.Tensor of bit strings n_shots x n_qubits.
    """

    return torch.concatenate(
        list(
            map(
                lambda bitstring: torch.broadcast_to(
                    bitstring_to_tensor(bitstring), [sample[bitstring], len(bitstring)]
                ),
                sample.keys(),
            )
        )
    )


def create_noise_matrix(
    noise_distribution: torch.distributions, n_shots: int, n_qubits: int
) -> Tensor:
    """
    A helper function that creates a noise matrix for bit string corruption.

    NB: The noise matrix is not square, as all bits are considered independent.

    Args:
        noise_distribution: Torch statistical distribution one of Gaussian,
        Uniform, or Poisson.
        n_shots: Number of shots/samples.
        n_qubits: Number of qubits

    Returns:
        A sample out of the requested distribution given the number of shots/samples.
    """
    # the noise_matrix should be available to the user if they want to do error correction
    return noise_distribution.sample([n_shots, n_qubits])


def bs_bitflip_corruption(
    err_idx: Tensor,
    sample: Tensor,
) -> Counter:
    """
    A function that incorporates the expected readout error in a sample of bit strings.

    given a noise matrix.

    Args:
        err_idx (Tensor): A Boolean array of bit string indices to be corrupted.
        sample (Tensor): A torch.Tensor of bit strings n_shots x n_qubits.

    Returns:
        Counter: A counter of bit strings after readout corruption.
    """

    corrupted = sample ^ err_idx

    return Counter([tensor_to_bitstring(k) for k in corrupted])


def bs_confusion_corruption(
    confusion_matrix: Tensor,
    sample: Tensor,
) -> Counter:
    """Given a confusion matrix and samples, corrupt by multinomial samples.

    Args:
        confusion_matrix (Tensor): Confusion matrix shape 2**n_bits x 2**n_bits.
        sample (Tensor): A torch.Tensor of bit strings n_shots x n_qubits.

    Returns:
        Counter: A counter of bit strings after readout corruption.
    """
    n_bits = sample.shape[1]
    sample_indices = (
        sample * (2 ** torch.arange(n_bits - 1, -1, -1, device=sample.device).long())
    ).sum(dim=1)
    corrupted_indices = torch.multinomial(
        confusion_matrix[sample_indices], num_samples=1, replacement=True
    ).squeeze()

    corrupted_bitstrings = torch.zeros_like(sample)
    for i in range(n_bits):
        corrupted_bitstrings[:, i] = (corrupted_indices // (2 ** (n_bits - i - 1))) % 2

    return Counter([tensor_to_bitstring(k) for k in corrupted_bitstrings])


def create_confusion_matrices(noise_matrix: Tensor, error_probability: float) -> Tensor:
    confusion_matrices = []
    for i in range(noise_matrix.size()[1]):
        column_tensor = noise_matrix[:, i]
        flip_proba = (
            flip_proba.mean().item()
            if len(flip_proba := column_tensor[column_tensor < error_probability]) > 0
            else 0.0
        )
        confusion_matrix = torch.tensor(
            [[1.0 - flip_proba, flip_proba], [flip_proba, 1.0 - flip_proba]],
            dtype=torch.float64,
        )
        confusion_matrices.append(confusion_matrix)
    return torch.stack(confusion_matrices)


class ReadoutInterface(ABC):
    """Interface for readout protocols."""

    @singledispatchmethod
    def apply(self, inputs, n_shots):
        """Apply protocol on an input to corrupt.
            Can be a tensor of probabilities or a list of counters.

        Args:
            inputs: Inputs (probabilities or list of counters) to corrupt
            n_shots: Number of shots
        """
        raise NotImplementedError


class ReadoutNoise(ReadoutInterface):
    """Simulate errors when sampling from a circuit.

    The model is simple as all bits are considered independent
    and are corrupted with an equal `error_probability`.

    The simulation is done by drawing samples from a `noise_distribution`.
    These samples are then compared to `error_probability` to specify
    which bits are corrupted.


    Attributes:
        n_qubits (int): Number of qubits.
        seed (int | None, optional): Random seed value. Defaults to None.
        error_probability (float, Tensor, optional): Error probabilities of wrong
              readout. Defaults to 0.1 at any position in the bit strings for every bit if None.
              If float, the same probability is applied to every bit. If a 1D tensor of size
              n_qubits, a different probability is set for each qubit.
              If a tensor of shape (n_qubits, 2, 2) is passed, that is a confusion matrix,
              we extract the error_probability
              and do not compute the confusion as in the other cases.
        noise_distribution (str, optional): Noise distribution type. Defaults to WhiteNoise.UNIFORM.
    """

    def __init__(
        self,
        n_qubits: int,
        error_probability: float | Tensor = 0.1,
        seed: int | None = None,
        noise_distribution: torch.distributions | None = WhiteNoise.UNIFORM,
    ) -> None:
        """Initializes ReadoutNoise.

        Args:
            n_qubits (int): Number of qubits.
            seed (int | None, optional): Random seed value. Defaults to None.
            error_probability (float, Tensor, optional): Error probabilities of wrong
              readout. Defaults to 0.1 at any position in the bit strings for every bit if None.
              If float, the same probability is applied to every bit. If a 1D tensor of size
              n_qubits, a different probability is set for each qubit.
              If a tensor of shape (n_qubits, 2, 2) is passed, that is a confusion matrix,
              we extract the error_probability
              and do not compute the confusion as in the other cases.
            noise_distribution (str, optional): Noise distribution type.
              Defaults to WhiteNoise.UNIFORM.
            confusion_matrix (Tensor, optional): The confusion matrix if available.

        """
        self.n_qubits = n_qubits
        self.seed = seed
        size_error_probability = (1,)
        if isinstance(error_probability, float):
            error_probability = torch.tensor(error_probability)

        elif isinstance(error_probability, Tensor):
            size_error_probability = tuple(error_probability.size())
            if (
                len(size_error_probability) == 1
                and len(error_probability) != self.n_qubits
            ):
                raise ValueError(
                    f"`error_probability` should have {n_qubits} elements."
                )
            if (len(size_error_probability) == 3) and (
                size_error_probability != (self.n_qubits, 2, 2)
            ):
                raise ValueError(
                    f"`error_probability` should have {(n_qubits, 2, 2)} elements."
                )
        else:
            raise ValueError(
                f"`error_probability` should be float, 1D or {(n_qubits, 2, 2)} tensor."
            )

        self.error_probability = error_probability
        self.noise_distribution = noise_distribution
        self._compute_confusion: bool = (
            False if size_error_probability == (self.n_qubits, 2, 2) else True
        )

        if not self._compute_confusion:
            self.confusion_matrix = self.error_probability
            self.error_probability = torch.empty((self.n_qubits, 2))
            self.error_probability[:, 0] = self.confusion_matrix[:, 0, 1]
            self.error_probability[:, 1] = self.confusion_matrix[:, 1, 0]
        else:
            self.confusion_matrix = torch.empty((self.n_qubits, 2, 2))

    def create_noise_matrix(self, n_shots: int) -> Tensor:
        """Create a noise matrix from a noise distribution.

        Also save the confusion matrix.

        Args:
            n_shots (int): Number of shots.

        Returns:
            Tensor | tuple[Tensor]: The noise matrix and possibly the confusion ones too.
        """

        if self.seed is not None:
            torch.manual_seed(self.seed)

        noise_matrix = create_noise_matrix(
            self.noise_distribution, n_shots, self.n_qubits
        )
        if self._compute_confusion:
            confusion_matrices = create_confusion_matrices(
                noise_matrix=noise_matrix, error_probability=self.error_probability
            )
            self.confusion_matrix = confusion_matrices
        return noise_matrix

    @singledispatchmethod
    def apply(self, inputs, n_shots):
        raise NotImplementedError

    @apply.register
    def _(self, inputs: Tensor, n_shots: int) -> Tensor:
        """Apply confusion matrix on probabilities.

        Args:
            inputs (Tensor): Batch of probability vectors.
            n_shots (int, optional): Number of shots.

        Returns:
            Tensor: Corrupted probabilities.
        """

        # Create binary representations
        n_states = inputs.shape[1]

        # Create binary representation of all states
        state_indices = torch.arange(n_states, device=inputs.device)
        binary_repr = (
            state_indices.unsqueeze(1)
            >> torch.arange(self.n_qubits - 1, -1, -1, device=inputs.device)
        ) & 1

        # Get input and output bits for all qubits at once
        input_bits = binary_repr.unsqueeze(0).expand(n_states, -1, -1)
        output_bits = binary_repr.unsqueeze(1).expand(-1, n_states, -1)

        # Get transition probabilities for each bit position
        self.create_noise_matrix(n_shots)
        confusion_matrices = self.confusion_matrix

        # Index into confusion matrix for all qubits at once
        # Shape: (n_states_out, n_states_in, n_qubits)
        qubit_transitions = confusion_matrices[
            torch.arange(self.n_qubits, device=inputs.device),
            output_bits,
            input_bits,
        ]
        transition_matrix = torch.prod(qubit_transitions, dim=-1)
        output_probs = torch.matmul(inputs, transition_matrix.T)
        return output_probs

    @apply.register
    def _(self, inputs: list, n_shots: int) -> list[Counter]:
        """Apply readout on counters represented as Counters.

        Args:
            inputs (list[Counter | OrderedCounter]): Samples of bit string as Counters.
            n_shots (int, optional): Number of shots to sample. Defaults to 1000.

        Returns:
            list[Counter]: Samples of corrupted bit strings
        """
        noise_matrix = self.create_noise_matrix(n_shots)
        err_idx = torch.as_tensor(noise_matrix < self.error_probability)

        corrupted_bitstrings = []
        for counter in inputs:
            sample = sample_to_matrix(counter)
            corrupted_bitstrings.append(
                bs_bitflip_corruption(err_idx=err_idx, sample=sample)
            )
        return corrupted_bitstrings


class CorrelatedReadoutNoise(ReadoutInterface):
    def __init__(
        self,
        confusion_matrix: Tensor,
        seed: int | None = None,
    ) -> None:
        """Initializes CorrelatedReadoutNoise.

        Args:
            confusion_matrices (Tensor): Confusion matrices of size (2**n_qubits, 2**n_qubits).
        """
        if (len(confusion_matrix.size()) != 2) or (
            confusion_matrix.size(0) != confusion_matrix.size(1)
        ):
            raise ValueError("The confusion matrix should be square")
        self.confusion_matrix = confusion_matrix
        self.n_qubits = int(log(confusion_matrix.size(0), 2))
        self.seed = seed

    @singledispatchmethod
    def apply(self, inputs, n_shots):
        raise NotImplementedError

    @apply.register
    def _(self, inputs: Tensor, n_shots: int) -> Tensor:
        """Apply confusion matrix on probabilities.

        Args:
            inputs (Tensor): Batch of probability vectors.
            n_shots (int, optional): Number of shots.

        Returns:
            Tensor: Corrupted probabilities.
        """
        output_probs = inputs @ self.confusion_matrix.T
        return output_probs

    @apply.register
    def _(self, inputs: list, n_shots: int) -> list[Counter]:
        """Apply readout on counters represented as Counters.

        Args:
            inputs (list[Counter |  OrderedCounter]): Samples of bit string as Counters.
            n_shots (int, optional): Number of shots to sample. Defaults to 1000.

        Returns:
            list[Counter]: Samples of corrupted bit strings
        """
        if self.seed is not None:
            torch.manual_seed(self.seed)
        corrupted_bitstrings = []
        for counter in inputs:
            sample = sample_to_matrix(counter)
            corrupted_bitstrings.append(
                bs_confusion_corruption(self.confusion_matrix, sample)
            )
        return corrupted_bitstrings



from __future__ import annotations

import pytest
import torch
from helpers import random_pauli_hamiltonian

import pyqtorch as pyq
from pyqtorch import DiffMode, expectation
from pyqtorch.matrices import COMPLEX_TO_REAL_DTYPES
from pyqtorch.primitives import Primitive
from pyqtorch.utils import (
    ATOL,
    GRADCHECK_ATOL,
    GRADCHECK_ATOL_hamevo,
    density_mat,
)


def circuit_example(
    n_qubits: int, n_layers: int
) -> tuple[pyq.QuantumCircuit, dict[str, torch.Tensor]]:
    rx = pyq.RX(0, param_name="theta_0")
    cry = pyq.CPHASE(0, 1, param_name="theta_1")
    rz = pyq.RZ(2, param_name="theta_2")

    cnot = pyq.CNOT(1, 2)
    ops = [rx, cry, rz, cnot] * n_layers
    circ = pyq.QuantumCircuit(n_qubits, ops)

    theta_0_value = torch.pi / 2
    theta_1_value = torch.pi
    theta_2_value = torch.pi / 4
    values = {
        "theta_0": torch.tensor([theta_0_value], requires_grad=True),
        "theta_1": torch.tensor([theta_1_value], requires_grad=True),
        "theta_2": torch.tensor([theta_2_value], requires_grad=True),
    }
    return circ, values


@pytest.mark.parametrize("n_qubits", [3, 4, 5])
@pytest.mark.parametrize("n_layers", [1, 2, 3])
def test_ad_observable(
    n_qubits: int,
    n_layers: int,
) -> None:
    circ, values_circuit = circuit_example(n_qubits, n_layers)
    state = pyq.zero_state(n_qubits)
    obs = pyq.Observable(pyq.RZ(0, "theta_obs"))
    assert obs.is_parametric
    assert obs.params == ["theta_obs"]
    values_obs = {"theta_obs": torch.tensor([torch.pi / 4], requires_grad=True)}
    values_all = values_circuit | values_obs

    exp_ad = expectation(circ, state, values_all, obs, DiffMode.AD)
    grad_ad = torch.autograd.grad(
        exp_ad, tuple(values_all.values()), torch.ones_like(exp_ad)
    )
    assert len(grad_ad) == len(values_circuit) + 1

    values_separated = {"circuit": values_circuit, "observables": values_obs}
    exp_ad_separate = expectation(circ, state, values_separated, obs, DiffMode.AD)
    assert torch.allclose(exp_ad_separate, exp_ad)
    grad_ad_obs = torch.autograd.grad(
        exp_ad_separate, tuple(values_obs.values()), torch.ones_like(exp_ad)
    )
    assert len(grad_ad_obs) == len(obs.params)
    assert torch.allclose(grad_ad[-1], grad_ad_obs[0])


@pytest.mark.parametrize("n_qubits", [3, 4, 5])
@pytest.mark.parametrize("n_layers", [1, 2, 3])
def test_adjoint_diff(n_qubits: int, n_layers: int) -> None:
    circ, values_ad = circuit_example(n_qubits, n_layers)
    obs = pyq.Observable(pyq.Z(0))
    assert not obs.is_parametric

    state = pyq.zero_state(n_qubits)
    values_adjoint = values_ad.copy()

    exp_ad = expectation(circ, state, values_ad, obs, DiffMode.AD)
    exp_adjoint = expectation(circ, state, values_adjoint, obs, DiffMode.ADJOINT)

    grad_ad = torch.autograd.grad(
        exp_ad, tuple(values_ad.values()), torch.ones_like(exp_ad)
    )
    assert len(grad_ad) == len(values_ad)

    grad_adjoint = torch.autograd.grad(
        exp_adjoint, tuple(values_adjoint.values()), torch.ones_like(exp_adjoint)
    )

    assert len(grad_ad) == len(grad_adjoint)
    for i in range(len(grad_ad)):
        if grad_adjoint[i] is not None:
            assert torch.allclose(grad_ad[i], grad_adjoint[i], atol=GRADCHECK_ATOL)

    # TODO higher order adjoint is not yet supported.
    # gradgrad_adjoint = torch.autograd.grad(
    #     grad_adjoint, tuple(values_adjoint.values()), torch.ones_like(grad_adjoint)
    # )


@pytest.mark.parametrize("n_qubits", [3, 4, 5])
@pytest.mark.parametrize("n_layers", [1, 2, 3])
@pytest.mark.parametrize("diagonal", [True, False])
@pytest.mark.parametrize("commuting_terms", [False, True])
def test_adjoint_diff_hamevo(
    n_qubits: int, n_layers: int, diagonal: bool, commuting_terms: bool
) -> None:

    cnot = pyq.CNOT(1, 2)
    ham = random_pauli_hamiltonian(
        n_qubits,
        k_1q=n_qubits,
        k_2q=0,
        default_scale_coeffs=1.0,
        diagonal=diagonal,
        commuting_terms=commuting_terms,
    )[0]
    ham_op = pyq.HamiltonianEvolution(ham, "t", qubit_support=tuple(range(n_qubits)))
    ops = [ham_op, cnot] * n_layers
    circ = pyq.QuantumCircuit(n_qubits, ops)
    obs = pyq.Observable(pyq.Z(0))

    t = torch.pi / 3

    state = pyq.zero_state(n_qubits)

    t_ad = torch.tensor([t], requires_grad=True)
    t_adjoint = torch.tensor([t], requires_grad=True)

    values_ad = {"t": t_ad}
    values_adjoint = {
        "t": t_adjoint,
    }
    exp_ad = expectation(circ, state, values_ad, obs, DiffMode.AD)
    exp_adjoint = expectation(circ, state, values_adjoint, obs, DiffMode.ADJOINT)

    grad_ad = torch.autograd.grad(
        exp_ad, tuple(values_ad.values()), torch.ones_like(exp_ad)
    )

    grad_adjoint = torch.autograd.grad(
        exp_adjoint, tuple(values_adjoint.values()), torch.ones_like(exp_adjoint)
    )

    assert len(grad_ad) == len(grad_adjoint)
    for i in range(len(grad_ad)):
        assert torch.allclose(grad_ad[i], grad_adjoint[i], atol=GRADCHECK_ATOL_hamevo)


@pytest.mark.parametrize("n_qubits", [3, 5])
@pytest.mark.parametrize("batch_size", [1, 2])
@pytest.mark.parametrize("ops_op", [pyq.Z, pyq.Y])
@pytest.mark.parametrize("dtype", [torch.complex64, torch.complex128])
def test_sampled_diff(
    n_qubits: int,
    batch_size: int,
    ops_op: Primitive,
    dtype: torch.dtype,
) -> None:
    rx = pyq.RX(0, param_name="theta_0")
    cry = pyq.CPHASE(0, 1, param_name="theta_1")
    rz = pyq.RZ(2, param_name="theta_2")
    cnot = pyq.CNOT(1, 2)
    ops = [rx, cry, rz, cnot]
    circ = pyq.QuantumCircuit(n_qubits, ops).to(dtype)
    obs = pyq.Observable(pyq.Add([ops_op(i) for i in range(n_qubits)])).to(dtype)

    theta_0_value = torch.pi / 2
    theta_1_value = torch.pi
    theta_2_value = torch.pi / 4

    state = pyq.random_state(n_qubits, batch_size=batch_size, dtype=dtype)
    theta_0_ad = torch.tensor(
        [theta_0_value], requires_grad=True, dtype=COMPLEX_TO_REAL_DTYPES[dtype]
    )
    theta_1_ad = torch.tensor(
        [theta_1_value], requires_grad=True, dtype=COMPLEX_TO_REAL_DTYPES[dtype]
    )
    theta_2_ad = torch.tensor(
        [theta_2_value], requires_grad=True, dtype=COMPLEX_TO_REAL_DTYPES[dtype]
    )
    values = {"theta_0": theta_0_ad, "theta_1": theta_1_ad, "theta_2": theta_2_ad}

    exp_ad = expectation(circ, state, values, obs, DiffMode.AD)
    exp_ad_sampled = expectation(
        circ,
        state,
        values,
        obs,
        DiffMode.AD,
        n_shots=100000,
    )
    assert torch.allclose(exp_ad, exp_ad_sampled, atol=1e-01)

    # test density mat
    dm = density_mat(state)
    exp_ad_dm = expectation(circ, dm, values, obs, DiffMode.AD)
    exp_ad_sampled_dm = expectation(
        circ,
        dm,
        values,
        obs,
        DiffMode.AD,
        n_shots=100000,
    )
    assert torch.allclose(exp_ad_dm, exp_ad, atol=ATOL)
    assert torch.allclose(exp_ad, exp_ad_sampled_dm, atol=1e-01)


@pytest.mark.skip  # Adjoint Scale is currently not supported
@pytest.mark.parametrize("dtype", [torch.complex64, torch.complex128])
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("n_qubits", [2])
def test_adjoint_scale(dtype: torch.dtype, batch_size: int, n_qubits: int) -> None:
    ops = [pyq.Scale(pyq.X(0), "theta_4")]

    theta_4_value = torch.rand(1, dtype=dtype)
    circ = pyq.QuantumCircuit(n_qubits, ops).to(dtype)

    state = pyq.random_state(n_qubits, batch_size, dtype=dtype)

    theta_4_ad = torch.tensor([theta_4_value], requires_grad=True)
    theta_4_adjoint = torch.tensor([theta_4_value], requires_grad=True)

    values_ad = torch.nn.ParameterDict(
        {
            "theta_4": theta_4_ad,
        }
    ).to(COMPLEX_TO_REAL_DTYPES[dtype])
    values_adjoint = torch.nn.ParameterDict(
        {
            "theta_4": theta_4_adjoint,
        }
    ).to(COMPLEX_TO_REAL_DTYPES[dtype])

    obs = pyq.Observable(pyq.Z(0)).to(dtype)
    exp_ad = expectation(circ, state, values_ad, obs, DiffMode.AD)
    exp_adjoint = expectation(circ, state, values_adjoint, obs, DiffMode.ADJOINT)

    grad_ad = torch.autograd.grad(
        exp_ad, tuple(values_ad.values()), torch.ones_like(exp_ad)
    )

    grad_adjoint = torch.autograd.grad(
        exp_adjoint, tuple(values_adjoint.values()), torch.ones_like(exp_adjoint)
    )

    assert len(grad_ad) == len(grad_adjoint)
    for i in range(len(grad_ad)):
        assert torch.allclose(grad_ad[i], grad_adjoint[i], atol=GRADCHECK_ATOL)


# Note pyq does not support using multiple times the same angle
@pytest.mark.parametrize("n_qubits", [3, 4, 5])
@pytest.mark.parametrize("batch_size", [1, 10])
@pytest.mark.parametrize("op_obs", [pyq.Z, pyq.X, pyq.Y])
@pytest.mark.parametrize("dtype", [torch.complex64, torch.complex128])
@pytest.mark.parametrize("different_names", [False, True])
def test_all_diff_singlegap(
    n_qubits: int,
    batch_size: int,
    op_obs: Primitive,
    dtype: torch.dtype,
    different_names: bool,
) -> None:
    name_angles = "theta"
    dtype_val = COMPLEX_TO_REAL_DTYPES[dtype]

    if different_names:
        ops_rx = pyq.Sequence(
            [
                pyq.RX(i, param_name=name_angles + "_x_" + str(i))
                for i in range(n_qubits)
            ]
        )
        ops_rz = pyq.Sequence(
            [
                pyq.RZ(i, param_name=name_angles + "_z_" + str(i))
                for i in range(n_qubits)
            ]
        )
        values = {
            name_angles
            + "_x_"
            + str(i): torch.rand(1, dtype=dtype_val, requires_grad=True)
            for i in range(n_qubits)
        }
        values.update(
            {
                name_angles
                + "_z_"
                + str(i): torch.rand(1, dtype=dtype_val, requires_grad=True)
                for i in range(n_qubits)
            }
        )
    else:
        ops_rx = pyq.Sequence(
            [pyq.RX(i, param_name=name_angles + "_x") for i in range(n_qubits)]
        )
        ops_rz = pyq.Sequence(
            [pyq.RZ(i, param_name=name_angles + "_z") for i in range(n_qubits)]
        )
        values = {
            name_angles + "_x": torch.rand(1, dtype=dtype_val, requires_grad=True),
            name_angles + "_z": torch.rand(1, dtype=dtype_val, requires_grad=True),
        }
    cnot = pyq.CNOT(1, 2)
    ops = [ops_rx, ops_rz, cnot]

    circ = pyq.QuantumCircuit(n_qubits, ops).to(dtype)
    obs = pyq.Observable([op_obs(i) for i in range(n_qubits)]).to(dtype)
    state = pyq.random_state(n_qubits, batch_size, dtype=dtype)

    exp_ad = expectation(circ, state, values, obs, DiffMode.AD)
    exp_adjoint = expectation(circ, state, values, obs, DiffMode.ADJOINT)
    exp_gpsr = expectation(circ, state, values, obs, DiffMode.GPSR)

    assert torch.allclose(exp_ad, exp_adjoint)
    assert torch.allclose(exp_ad, exp_gpsr)

    grad_ad = torch.autograd.grad(
        exp_ad, tuple(values.values()), torch.ones_like(exp_ad), create_graph=True
    )

    grad_adjoint = torch.autograd.grad(
        exp_adjoint,
        tuple(values.values()),
        torch.ones_like(exp_adjoint),
        create_graph=True,
    )

    grad_gpsr = torch.autograd.grad(
        exp_gpsr, tuple(values.values()), torch.ones_like(exp_gpsr), create_graph=True
    )

    # check first order gradients
    assert len(grad_ad) == len(grad_adjoint) == len(grad_gpsr)
    for i in range(len(grad_ad)):
        assert torch.allclose(
            grad_ad[i], grad_adjoint[i], atol=GRADCHECK_ATOL
        ) and torch.allclose(grad_ad[i], grad_gpsr[i], atol=GRADCHECK_ATOL)

    # TODO higher order adjoint is not yet supported.
    # gradgrad_adjoint = torch.autograd.grad(
    #     grad_adjoint, tuple(values.values()), torch.ones_like(grad_adjoint)
    # )

    for i in range(len(grad_ad)):
        gradgrad_ad = torch.autograd.grad(
            grad_ad[i],
            tuple(values.values()),
            torch.ones_like(grad_ad[i]),
            create_graph=True,
        )

        gradgrad_gpsr = torch.autograd.grad(
            grad_gpsr[i],
            tuple(values.values()),
            torch.ones_like(grad_gpsr[i]),
            create_graph=True,
        )

        assert len(gradgrad_ad) == len(gradgrad_gpsr)

        # check second order gradients
        for j in range(len(gradgrad_ad)):
            assert torch.allclose(gradgrad_ad[j], gradgrad_gpsr[j], atol=GRADCHECK_ATOL)



from __future__ import annotations

from functools import reduce
from logging import getLogger
from operator import add
from typing import Any, Generator, NoReturn, Union

import numpy as np
import torch
from torch import Tensor, einsum, rand
from torch.nn import Module, ModuleList, ParameterDict

from pyqtorch.apply import apply_operator, apply_operator_dm
from pyqtorch.embed import ConcretizedCallable, Embedding
from pyqtorch.matrices import add_batch_dim
from pyqtorch.primitives import CNOT, RX, RY, I, Parametric, Primitive, X, Y, Z
from pyqtorch.utils import (
    DensityMatrix,
    Operator,
    State,
)

from .sequence import Sequence

BATCH_DIM = 2
PAULI_OPS = (X, Y, Z, I)

logger = getLogger(__name__)


class Scale(Sequence):
    """
    Generic container for multiplying a 'Primitive', 'Sequence' or 'Add' instance by a parameter.

    Attributes:
        operations: Operations as a Sequence, Add, or a single Primitive operation.
        param_name: Name of the parameter to multiply operations with.
    """

    def __init__(
        self,
        operations: Union[Primitive, Sequence, Add],
        param_name: str | float | int | Tensor | ConcretizedCallable,
    ):
        """
        Initializes a Scale object.

        Arguments:
            operations: Operations as a Sequence, Add, or a single Primitive operation.
            param_name: Name of the parameter to multiply operations with.
        """
        if not isinstance(operations, (Primitive, Sequence, Add)):
            raise ValueError("Scale only supports a single operation, Sequence or Add.")
        if not isinstance(param_name, (str, int, float, Tensor, ConcretizedCallable)):
            raise TypeError(
                "Only str, int, float, Tensor or ConcretizedCallable types \
                are supported for param_name"
            )
        self.is_diagonal = operations.is_diagonal
        super().__init__([operations])
        self.param_name = param_name

    def forward(
        self,
        state: Tensor,
        values: dict[str, Tensor] | ParameterDict | None = None,
        embedding: Embedding | None = None,
    ) -> State:
        """
        Apply the operation(s) multiplying by the parameter value.

        Arguments:
            state: Input state.
            values: Parameter value.

        Returns:
            The transformed state.
        """
        values = values or dict()
        if embedding is not None:
            values = embedding(values)

        if isinstance(self.param_name, str):
            scale = values[self.param_name]
        elif isinstance(self.param_name, Tensor):
            scale = self.param_name
        elif isinstance(self.param_name, ConcretizedCallable):
            scale = self.param_name(values)

        return scale * self.operations[0].forward(state, values)

    def tensor(
        self,
        values: dict[str, Tensor] | None = None,
        embedding: Embedding | None = None,
        full_support: tuple[int, ...] | None = None,
        diagonal: bool = False,
    ) -> Operator:
        """
        Get the corresponding unitary over n_qubits.

        Arguments:
            values: Parameter value.
            embedding: An optional embedding.
            full_support: Can be higher than the number of qubit support.

        Returns:
            The unitary representation.
        """
        values = values or dict()
        if embedding is not None:
            values = embedding(values)
        use_diagonal = diagonal and self.is_diagonal
        if isinstance(self.param_name, str):
            scale = values[self.param_name]
        elif isinstance(self.param_name, (Tensor, int, float)):
            scale = self.param_name
        elif isinstance(self.param_name, ConcretizedCallable):
            scale = self.param_name(values)

        return scale * self.operations[0].tensor(
            values, full_support=full_support, diagonal=use_diagonal
        )

    def flatten(self) -> list[Scale]:
        """This method should only be called in the AdjointExpectation,
        where the `Scale` is only supported for Primitive (and not Sequences)
        so we don't want to flatten this to preserve the scale parameter.

        Returns:
            The Scale within a list.
        """
        return [self]

    def to(self, *args: Any, **kwargs: Any) -> Scale:
        """Perform conversions for dtype or device.

        Returns:
            Converted Scale.
        """
        super().to(*args, **kwargs)
        if not isinstance(self.param_name, (str, float, int)):
            self.param_name = self.param_name.to(*args, **kwargs)

        return self


def disjoint_supports(
    operations: ModuleList | list[Module], qubit_support: tuple[int, ...]
) -> bool:
    """Check all operations are defined over disjoint supports.

    Args:
        operations (ModuleList): Operations.
        qubit_support (tuple[int, ...]): Support of all operations.

    Returns:
        bool: True if all operations are defined on independent supports.
    """
    qubit_supports_intersect = [set(op.qubit_support) for op in operations]

    disjoint_sets: set[int] = set()
    for s in qubit_supports_intersect:
        if s.isdisjoint(disjoint_sets):
            disjoint_sets.update(s)
        else:
            return False
    if len(disjoint_sets) == len(qubit_support):
        return True

    return False


def symplectic_inner_product(
    x1: np.array, z1: np.array, x2: np.array, z2: np.array
) -> int:
    """Sympletic inner product for checking commutation fot two pauli strings p1 and p2
    represented by their binary representation (`x1`, `z1`), (`x2, `z2``)

    Returns:
        int: 0 if the pauli strings commute, 1 if they anti-commute
    """
    dot1 = np.sum(x1 & z2) % 2
    dot2 = np.sum(z1 & x2) % 2
    return (dot1 ^ dot2).item()  # XOR of the two dot products


class Add(Sequence):
    """
    The 'add' operation applies all 'operations' to 'state' and returns the sum of states.

    Attributes:
        operations: List of operations to add up.
    """

    def __init__(self, operations: list[Module]):
        super().__init__(operations=operations)

    def forward(
        self,
        state: State,
        values: dict[str, Tensor] | ParameterDict | None = None,
        embedding: Embedding | None = None,
    ) -> State:
        """
        Apply the operations multiplying by the parameter values.

        Arguments:
            state: Input state.
            values: Parameter value.

        Returns:
            The transformed state.
        """
        values = values or dict()
        return reduce(add, (op(state, values, embedding) for op in self.operations))

    def tensor(
        self,
        values: dict | None = None,
        embedding: Embedding | None = None,
        full_support: tuple[int, ...] | None = None,
        diagonal: bool = False,
    ) -> Tensor:
        """
        Get the corresponding sum of unitaries over n_qubits.

        Arguments:
            values: Parameter value.
            Can be higher than the number of qubit support.


        Returns:
            The unitary representation.
        """
        values = values or dict()
        use_diagonal = diagonal and self.is_diagonal
        if full_support is None:
            full_support = self.qubit_support
        elif not set(self.qubit_support).issubset(set(full_support)):
            raise ValueError(
                "Expanding tensor operation requires a `full_support` argument "
                "larger than or equal to the `qubit_support`."
            )
        if not use_diagonal:
            mat = torch.zeros(
                (2 ** len(full_support), 2 ** len(full_support), 1), device=self.device
            )
        else:
            mat = torch.zeros((2 ** len(full_support), 1), device=self.device)
        return reduce(
            add,
            (
                op.tensor(values, embedding, full_support, diagonal=use_diagonal)
                for op in self.operations
            ),
            mat,
        )

    def _symplectic_commute(self) -> bool:
        """A predicate to check if operator is composed of communting pauli strings.

        Achieved by computing the symplectic inner product from string-based representations.

        Reference:
            S. Aaronson, D. Gottesman, Improved Simulation of Stabilizer Circuits,
            Phys. Rev. A 70, 052328 (2004). arXiv:quant-ph/0406196
        """

        if all(isinstance(op, PAULI_OPS) for op in self._flatten()):
            qubit_support = self.qubit_support
            n_qubits = len(qubit_support)
            n_strings = len(self.operations)

            # build binary representation of pauli operations for symplectic product
            binary_vectors = [
                (np.zeros(n_qubits, dtype=int), np.zeros(n_qubits, dtype=int))
                for _ in range(n_strings)
            ]
            for ind_op, op in enumerate(self.operations):
                if isinstance(op, Sequence):
                    subops = op._flatten()
                    if not disjoint_supports(subops, op.qubit_support):
                        return False
                else:
                    subops = [
                        op,
                    ]
                for subop in subops:
                    qubit = qubit_support.index(subop.qubit_support[0])
                    if isinstance(subop, X):
                        binary_vectors[ind_op][0][qubit] = 1
                        binary_vectors[ind_op][1][qubit] = 1
                    if isinstance(subop, Y):
                        binary_vectors[ind_op][0][qubit] = 1
                    if isinstance(subop, Z):
                        binary_vectors[ind_op][1][qubit] = 1

            # Check commutation using symplectic inner product

            for i in range(n_strings):
                for j in range(i + 1, n_strings):
                    product = symplectic_inner_product(
                        *binary_vectors[i], *binary_vectors[j]
                    )

                    if product == 1:
                        return False  # Found a pair that doesn't commute

            return True

        return False

    def commuting_terms(self) -> bool:
        """Predicate for all operations being composed of commuting terms.

        Useful for faster HamiltonianEvolution forward methods.
        """

        if len(self.operations) == 1:
            return False

        if (
            disjoint_supports(self.operations, self.qubit_support)
            or self._symplectic_commute()
        ):
            return True

        return False


class Merge(Sequence):
    def __init__(
        self,
        operations: list[Module],
    ):
        """
        Merge a sequence of single qubit operations acting on the same qubit into a single
        einsum operation.

        Arguments:
            operations: A list of single qubit operations.

        """
        if (
            isinstance(operations, (list, ModuleList))
            and all([isinstance(op, (Primitive, Parametric)) for op in operations])
            and all(list([len(op.qubit_support) == 1 for op in operations]))
            and len(list(set([op.qubit_support[0] for op in operations]))) == 1
        ):
            # We want all operations to act on the same qubit

            super().__init__(operations)
            self.qubits = operations[0].qubit_support
        else:
            raise TypeError(
                f"Require all operations to act on a single qubit. Got: {operations}."
            )

        self._contains_noise = sum([op.noise is not None for op in self.operations])

    def _flatten(self) -> ModuleList:
        """This method avoids returning individual operations when checking commutation."""
        return ModuleList([self])

    def forward(
        self,
        state: Tensor,
        values: dict[str, Tensor] | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        batch_size = state.shape[-1]
        values = values or dict()
        if values:
            batch_size = max(
                batch_size,
                max(
                    list(
                        map(
                            lambda t: (
                                len(t) if isinstance(t, Tensor) and t.ndim > 0 else 1
                            ),
                            values.values(),
                        )
                    )
                ),
            )

        if self._contains_noise:
            # noisy cannot use merged in tensors, fall back to super forward
            return super().forward(state, values, embedding)

        if isinstance(state, DensityMatrix):
            return apply_operator_dm(
                state,
                add_batch_dim(self.tensor(values, embedding), batch_size),
                self.qubits,
            )

        return apply_operator(
            state,
            add_batch_dim(self.tensor(values, embedding), batch_size),
            self.qubits,
        )

    def tensor(
        self,
        values: dict[str, Tensor] | None = None,
        embedding: Embedding | None = None,
        full_support: tuple[int, ...] | None = None,
        diagonal: bool = False,
    ) -> Tensor:
        # We reverse the list of tensors here since matmul is not commutative.
        values = values or dict()
        use_diagonal = diagonal and self.is_diagonal
        if not use_diagonal:
            return reduce(
                lambda u0, u1: einsum("ijb,jkb->ikb", u0, u1),
                (
                    op.tensor(values, embedding, full_support)
                    for op in reversed(self.operations)
                ),
            )
        else:
            return reduce(
                lambda u0, u1: einsum("jb,jb->jb", u0, u1),
                (
                    op.tensor(values, embedding, full_support, diagonal=True)
                    for op in reversed(self.operations)
                ),
            )


def hea(n_qubits: int, depth: int, param_name: str) -> tuple[ModuleList, ParameterDict]:
    def _idx() -> Generator[int, Any, NoReturn]:
        i = 0
        while True:
            yield i
            i += 1

    def idxer() -> Generator[int, Any, None]:
        yield from _idx()

    idx = idxer()
    ops = []
    for _ in range(depth):
        layer = []
        for i in range(n_qubits):
            layer += [
                Merge([fn(i, f"{param_name}_{next(idx)}") for fn in [RX, RY, RX]])
            ]
        ops += layer
        ops += [
            Sequence([CNOT(i % n_qubits, (i + 1) % n_qubits) for i in range(n_qubits)])
        ]
    params = ParameterDict(
        {f"{param_name}_{n}": rand(1, requires_grad=True) for n in range(next(idx))}
    )
    return ops, params



from __future__ import annotations

import torch
from torch import Tensor

DEFAULT_REAL_DTYPE = torch.float64
DEFAULT_MATRIX_DTYPE = torch.cdouble

IMAT = torch.eye(2, dtype=DEFAULT_MATRIX_DTYPE)
XMAT = torch.tensor([[0, 1], [1, 0]], dtype=DEFAULT_MATRIX_DTYPE)
YMAT = torch.tensor([[0, -1j], [1j, 0]], dtype=DEFAULT_MATRIX_DTYPE)
ZMAT = torch.tensor([[1, 0], [0, -1]], dtype=DEFAULT_MATRIX_DTYPE)
SMAT = torch.tensor([[1, 0], [0, 1j]], dtype=DEFAULT_MATRIX_DTYPE)
SDAGGERMAT = torch.tensor([[1, 0], [0, -1j]], dtype=DEFAULT_MATRIX_DTYPE)
TMAT = torch.tensor(
    [
        [1, 0],
        [0, torch.exp(torch.tensor(1.0j * torch.pi / 4, dtype=DEFAULT_MATRIX_DTYPE))],
    ],
    dtype=DEFAULT_MATRIX_DTYPE,
)
NMAT = torch.tensor([[0, 0], [0, 1]], dtype=DEFAULT_MATRIX_DTYPE)
IDIAG = torch.tensor([1, 1], dtype=DEFAULT_MATRIX_DTYPE)

SWAPMAT = torch.tensor(
    [[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]], dtype=DEFAULT_MATRIX_DTYPE
)
CSWAPMAT = torch.tensor(
    [
        [1, 0, 0, 0, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0, 0, 0],
        [0, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 1, 0],
        [0, 0, 0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 1],
    ],
    dtype=DEFAULT_MATRIX_DTYPE,
)
HMAT = (
    1
    / torch.sqrt(torch.tensor(2.0, dtype=DEFAULT_MATRIX_DTYPE))
    * torch.tensor([[1, 1], [1, -1]], dtype=DEFAULT_MATRIX_DTYPE)
)


def PROJMAT(ket: Tensor, bra: Tensor) -> Tensor:
    return torch.outer(ket, bra)


OPERATIONS_DICT = {
    "I": IMAT,
    "X": XMAT,
    "Y": YMAT,
    "Z": ZMAT,
    "S": SMAT,
    "SDAGGER": SDAGGERMAT,
    "T": TMAT,
    "N": NMAT,
    "PROJ": PROJMAT,
    "H": HMAT,
    "SWAP": SWAPMAT,
    "CSWAP": CSWAPMAT,
}


def parametric_unitary(
    theta: torch.Tensor,
    P: torch.Tensor,
    identity_mat: torch.Tensor,
    batch_size: int,
    diagonal: bool = False,
    a: float = 0.5,  # noqa: E741
) -> torch.Tensor:
    """Compute the exponentiation of a Pauli matrix :math:`P`

    The exponentiation is given by:
    :math:`exp(-i a \\theta P ) = I cos(r \\theta) - i a P sin(r \\theta) / r`

    where :math:`a` is a prefactor
    and :math:`r = a * sg / 2`, :math:`sg` corresponding to the spectral gap.

    Here, we assume :math:`sg = 2`

    Args:
        theta (torch.Tensor): Parameter values.
        P (torch.Tensor): Pauli matrix to exponentiate.
        identity_mat (torch.Tensor): Identity matrix
        batch_size (int): Batch size of parameters.
        diagonal (bool): if dealing with diagonal operation.
        a (float): Prefactor.

    Returns:
        torch.Tensor: The exponentiation of P
    """
    cos_t = (
        torch.cos(theta * a)
        .unsqueeze(0)
        .to(dtype=identity_mat.dtype, device=identity_mat.device)
    )
    sin_t = (
        torch.sin(theta * a)
        .unsqueeze(0)
        .to(dtype=identity_mat.dtype, device=identity_mat.device)
    )
    batch_imat = identity_mat.unsqueeze(-1)
    batch_operation_mat = P.unsqueeze(-1)
    if not diagonal:
        cos_t = cos_t.unsqueeze(1)
        sin_t = sin_t.unsqueeze(1)
        cos_t = cos_t.repeat((2, 2, 1))
        sin_t = sin_t.repeat((2, 2, 1))

        batch_imat = batch_imat.repeat(1, 1, batch_size)
        batch_operation_mat = batch_operation_mat.repeat(1, 1, batch_size)
    else:
        batch_imat = batch_imat.repeat(1, batch_size)
        batch_operation_mat = batch_operation_mat.repeat(1, batch_size)

    return cos_t * batch_imat - 1j * sin_t * batch_operation_mat


def _dagger(matrices: torch.Tensor, diagonal: bool = False) -> torch.Tensor:
    return (
        torch.permute(matrices.conj(), (1, 0, 2)) if not diagonal else matrices.conj()
    )


def _jacobian(
    theta: torch.Tensor,
    P: torch.Tensor,
    identity_mat: torch.Tensor,
    batch_size: int,
    diagonal: bool = False,
) -> torch.Tensor:
    cos_t = torch.cos(theta / 2).unsqueeze(0)
    sin_t = torch.sin(theta / 2).unsqueeze(0)
    batch_imat = identity_mat.unsqueeze(-1)
    batch_operation_mat = P.unsqueeze(-1)
    if not diagonal:
        cos_t = cos_t.unsqueeze(1)
        sin_t = sin_t.unsqueeze(1)
        cos_t = cos_t.repeat((2, 2, 1))
        sin_t = sin_t.repeat((2, 2, 1))

        batch_imat = batch_imat.repeat(1, 1, batch_size)
        batch_operation_mat = batch_operation_mat.repeat(1, 1, batch_size)
    else:
        batch_imat = batch_imat.repeat(1, batch_size)
        batch_operation_mat = batch_operation_mat.repeat(1, batch_size)

    return -1 / 2 * (sin_t * batch_imat + 1j * cos_t * batch_operation_mat)


def controlled(
    operation: torch.Tensor,
    batch_size: int,
    n_control_qubits: int = 1,
    diagonal: bool = False,
) -> torch.Tensor:
    if diagonal:
        _controlled: torch.Tensor = (
            torch.ones(
                2 ** (n_control_qubits + 1),
                dtype=operation.dtype,
                device=operation.device,
            )
            .unsqueeze(1)
            .repeat(1, batch_size)
        )
        _controlled[2 ** (n_control_qubits + 1) - 2 :, :] = operation
    else:
        _controlled = (
            torch.eye(
                2 ** (n_control_qubits + 1),
                dtype=operation.dtype,
                device=operation.device,
            )
            .unsqueeze(2)
            .repeat(1, 1, batch_size)
        )
        _controlled[
            2 ** (n_control_qubits + 1) - 2 :, 2 ** (n_control_qubits + 1) - 2 :, :
        ] = operation
    return _controlled


COMPLEX_TO_REAL_DTYPES = {
    torch.complex128: torch.float64,
    torch.complex64: torch.float32,
    torch.complex32: torch.float16,
}


def add_batch_dim(operator: Tensor, batch_size: int = 1) -> Tensor:
    """In case we have a sequence of batched parametric gates mixed with primitive gates,
    we adjust the batch_dim of the primitive gates to match."""
    return (
        operator.repeat(1, 1, batch_size)
        if operator.shape != (2, 2, batch_size)
        else operator
    )



from __future__ import annotations

from dataclasses import dataclass

import torch
from torch import dtype


@dataclass
class AdaptiveSolverOptions:

    atol: float = 1e-8
    rtol: float = 1e-6
    max_steps: int = 100_000
    safety_factor: float = 0.9
    min_factor: float = 0.2
    max_factor: float = 5.0
    ctype: dtype = torch.complex128
    rtype: dtype = torch.float64
    use_sparse: bool = False


@dataclass
class KrylovSolverOptions:

    max_krylov: int = 80
    exp_tolerance: float = 1e-10
    norm_tolerance: float = 1e-10
    use_sparse: bool = False



from __future__ import annotations

from math import log, sqrt
from typing import Any, Callable

import torch
from torch import Tensor

from pyqtorch.matrices import XMAT, YMAT, ZMAT
from pyqtorch.qubit_support import Support
from pyqtorch.time_dependent.mesolve import mesolve
from pyqtorch.utils import (
    DensityMatrix,
    SolverType,
    density_mat,
    expand_operator,
    permute_basis,
)


class AnalogNoise(torch.nn.Module):
    """AnalogNoise is used within `HamiltonianEvolution`
        when solving the Schrödinger or a master (Lindblad) equation.

    Attributes:
        noise_operators (list[Tensor]): The list of jump operators
            to use with mesolve. Note you can only define them as
            2D tensors, so no batchsize.
        qubit_support (int | tuple[int, ...] | Support): The qubits
            where the noise_operators are defined over.

    """

    def __init__(
        self,
        noise_operators: list[Tensor],
        qubit_support: int | tuple[int, ...] | Support,
    ) -> None:
        super().__init__()
        self._qubit_support = (
            qubit_support
            if isinstance(qubit_support, Support)
            else Support(target=qubit_support)
        )

        for index, tensor in enumerate(noise_operators):
            if len(self._qubit_support) < int(log(tensor.shape[1], 2)):
                raise ValueError(
                    "Tensor dimensions should match the length of the qubit support."
                    f"Tensor {tensor} has incompatible dimensions {int(log(tensor.shape[1], 2))}."
                )
            self.register_buffer(f"noise_operators_{index}", tensor)

        self._device: torch.device = noise_operators[0].device
        self._dtype: torch.dtype = noise_operators[0].dtype

    @property
    def noise_operators(self) -> list[Tensor]:
        """Get noise_operators.

        Returns:
            list[Tensor]: Noise operators.
        """
        return [
            getattr(self, f"noise_operators_{i}") for i in range(len(self._buffers))
        ]

    @property
    def qubit_support(self) -> tuple[int, ...]:
        """Getter qubit_support.

        Returns:
            Support: Tuple of sorted qubits.
        """
        return self._qubit_support.sorted_qubits

    def extra_repr(self) -> str:
        return f"noise_operators: {self.noise_operators}"

    def _noise_operators(
        self,
        full_support: tuple[int, ...] | None = None,
    ) -> list[Tensor]:
        """Obtain noise operators expended with full support.

        Args:
            full_support (tuple[int, ...] | None, optional): _description_. Defaults to None.

        Returns:
            list[Tensor]: Noise operators defined over `full_support`.
        """
        list_ops = self.noise_operators
        if self._qubit_support.qubits != self.qubit_support:
            list_ops = [
                permute_basis(
                    blockmat.unsqueeze(-1), self._qubit_support.qubits, inv=True
                ).squeeze(-1)
                for blockmat in list_ops
            ]
        if full_support is None:
            return list_ops
        else:
            return [
                expand_operator(
                    blockmat.unsqueeze(-1), self.qubit_support, full_support
                ).squeeze(-1)
                for blockmat in list_ops
            ]

    def to(self, *args: Any, **kwargs: Any) -> AnalogNoise:
        """Do device or dtype conversions.

        Returns:
            AnalogNoise: Converted instance.
        """
        super().to(*args, **kwargs)
        self._device = self.noise_operators_0.device
        self._dtype = self.noise_operators_0.dtype
        return self

    def forward(
        self,
        state: Tensor,
        H: Callable[..., Any],
        tsave: list | Tensor,
        solver: SolverType,
        options: dict[str, Any] | None = None,
        full_support: tuple[int, ...] | None = None,
    ) -> Tensor:
        """Obtain the output density matrix by solving a Shrodinger equation.

        This uses the `mesolve` function.

        Args:
            state (Tensor): Input state or density matrix.
            H (Callable[..., Any]): Time-dependent hamiltonian fonction.
            tsave (list | Tensor): Tensor containing simulation time instants.
            solver (SolverType): Name of the solver to use.
            options (dict[str, Any] | None, optional): Additional options passed to the solver.
                Defaults to None.
            full_support (tuple[int, ...] | None, optional): The qubits the returned tensor
                will be defined over. Defaults to None for only using the qubit_support.

        Returns:
            Tensor: Output density matrix from solver.
        """
        if not isinstance(state, DensityMatrix):
            state = density_mat(state)
        sol = mesolve(
            H,
            state,
            self._noise_operators(full_support),
            tsave,
            solver,
            options=options,
        )
        return DensityMatrix(sol.states[-1])

    def __add__(self, other: AnalogNoise) -> AnalogNoise:
        return AnalogNoise(
            self.noise_operators + other.noise_operators,
            self._qubit_support + other._qubit_support,
        )


class AnalogDepolarizing(AnalogNoise):
    """
    Defines jump operators for an Analog Depolarizing channel.

    Under the depolarizing noise, a system in any state evolves
      to the maximally mixed state at a rate :math:`p`.

    The corresponding jump operators are :
    .. math::
        `L_{0,1,2} = \\sqrt{\\frac{p}{4}} \\sigma_{x,y,z}`

    where :math:`\\sigma_{x,y,z}` correspond to the unitaries of the X,Y,Z gates.

        Args:
            error_param (float): Rate of depolarizing.
            qubit_support (int | tuple[int, ...] | Support): Qubits defining the operation.
    """

    def __init__(
        self,
        error_param: float,
        qubit_support: int | tuple[int, ...] | Support,
    ) -> None:
        """Initializes AnalogDepolarizing.

        Args:
            error_param (float): Rate of depolarizing.
            qubit_support (int | tuple[int, ...] | Support): Qubits defining the operation.
        """
        coeff = sqrt(error_param / 4)
        L0 = coeff * XMAT.squeeze()
        L1 = coeff * YMAT.squeeze()
        L2 = coeff * ZMAT.squeeze()
        noise_operators: list[Tensor] = [L0, L1, L2]
        super().__init__(noise_operators, qubit_support)



from __future__ import annotations

from functools import cached_property
from typing import Any, Tuple
from uuid import uuid4

import torch
from torch import Tensor

from pyqtorch.embed import ConcretizedCallable, Embedding
from pyqtorch.matrices import (
    OPERATIONS_DICT,
    _jacobian,
    controlled,
    parametric_unitary,
)
from pyqtorch.noise import DigitalNoiseProtocol, _repr_noise
from pyqtorch.quantum_operation import QuantumOperation, Support

pauli_singleq_eigenvalues = torch.tensor([[-1.0], [1.0]], dtype=torch.cdouble)


class Parametric(QuantumOperation):
    """
    QuantumOperation taking parameters as input.

    Attributes:
        param_name: Name of parameters.
        parse_values: Method defining how to handle the values dictionary input.
    """

    n_params = 1

    def __init__(
        self,
        generator: str | Tensor,
        qubit_support: int | tuple[int, ...] | Support,
        param_name: str | int | float | torch.Tensor | ConcretizedCallable = "",
        noise: DigitalNoiseProtocol | None = None,
        diagonal: bool = False,
    ):
        """Initializes Parametric.

        Arguments:
            generator: Generator to use.
            qubit_support: Qubits to act on.
            param_name: Name of parameters.
            noise: Optional noise protocols to apply.
            diagonal: Whether the tensor generator is diagonal.
        """

        generator_operation = (
            OPERATIONS_DICT[generator] if isinstance(generator, str) else generator
        )
        if not isinstance(param_name, (str, int, float, Tensor, ConcretizedCallable)):
            raise TypeError(
                "Only str, int, float, Tensor or ConcretizedCallable types \
                are supported for param_name"
            )
        self.param_name = param_name
        self._param_uuid = str(uuid4())

        def parse_values(
            values: dict[str, Tensor] | Tensor | None = None,
            embedding: Embedding | None = None,
        ) -> Tensor:
            """The legacy way of using parametric gates:
               The Parametric gate received a string as a 'param_name' and performs a
               a lookup in the passed `values` dict for to retrieve the torch.Tensor passed
               under the key `param_name`.

            Arguments:
                values: A dict containing param_name:torch.Tensor pairs
                embedding: An optional embedding.
            Returns:
                A Torch Tensor denoting values for the `param_name`.
            """
            values = values or dict()
            if embedding is not None:
                values = embedding(values)
            # note: GPSR trick when the same param_name is used in many operations
            if self._param_uuid in values.keys():
                return Parametric._expand_values(values[self._param_uuid] + values[self.param_name])  # type: ignore[index]
            return Parametric._expand_values(values[self.param_name])  # type: ignore[index]

        def parse_tensor(
            values: dict[str, Tensor] | Tensor | None = None,
            embedding: Embedding | None = None,
        ) -> Tensor:
            """Functional version of the Parametric gate:
               In case the user did not pass a `param_name`,
               pyqtorch assumes `values` will be a torch.Tensor instead of a dict.

            Arguments:
                values: A dict containing param_name:torch.Tensor pairs
            Returns:
                A Torch Tensor with which to evaluate the Parametric Gate.
            """
            values = values or dict()
            # self.param_name will be ""
            return Parametric._expand_values(values)

        def parse_constant(
            values: dict[str, Tensor] | Tensor | None = None,
            embedding: Embedding | None = None,
        ) -> Tensor:
            """Fix a the parameter of a Parametric Gate to a numeric constant
               if the user passed a numeric input for the `param_name`.

            Arguments:
                values: A dict containing param_name:torch.Tensor pairs
            Returns:
                A Torch Tensor with which to evaluate the Parametric Gate.
            """
            values = values or dict()
            # self.param_name will be a torch.Tensor
            return Parametric._expand_values(
                torch.tensor(self.param_name, device=self.device, dtype=self.dtype)
            )

        def parse_concretized_callable(
            values: dict[str, Tensor] | Tensor | None = None,
            embedding: Embedding | None = None,
        ) -> Tensor:
            """Evaluate ConcretizedCallable object with given values.

            Arguments:
                values: A dict containing param_name:torch.Tensor pairs
            Returns:
                A Torch Tensor with which to evaluate the Parametric Gate.
            """
            # self.param_name will be a ConcretizedCallable
            values = values or dict()
            return Parametric._expand_values(self.param_name(values))  # type: ignore [operator]

        if param_name == "":
            self.parse_values = parse_tensor
            self.param_name = self.param_name
        elif isinstance(param_name, str):
            self.parse_values = parse_values
        elif isinstance(param_name, (float, int, torch.Tensor)):
            self.parse_values = parse_constant
        elif isinstance(param_name, ConcretizedCallable):
            self.parse_values = parse_concretized_callable

        # Parametric is defined by generator operation and a function
        # The function will use parsed parameter values to compute the unitary
        super().__init__(
            generator_operation,
            qubit_support,
            operator_function=self._construct_parametric_base_op,
            noise=noise,
            diagonal=diagonal if isinstance(generator, Tensor) else False,
        )
        self.register_buffer("identity", OPERATIONS_DICT["I"])

    def extra_repr(self) -> str:
        """String representation of the operation.

        Returns:
            String with information on operation.
        """
        return f"target: {self.qubit_support}, param: {self.param_name}" + _repr_noise(
            self.noise
        )

    @property
    def is_parametric(self) -> bool:
        return isinstance(self.param_name, str) and self.param_name != ""

    def __hash__(self) -> int:
        """Hash qubit support and param_name

        Returns:
            Hash value
        """
        return hash(self.qubit_support) + hash(self.param_name)

    @staticmethod
    def _expand_values(values: Tensor) -> Tensor:
        """Expand values if necessary.

        Arguments:
            values: Values of parameters
        Returns:
            Values of parameters expanded.

        """
        return values.unsqueeze(0) if len(values.size()) == 0 else values

    def _construct_parametric_base_op(
        self,
        values: dict[str, Tensor] | Tensor | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        """
        Get the corresponding unitary with parsed values.

        Arguments:
            values: A dict containing a Parameter name and value.
            embedding: An optional embedding for parameters.

        Returns:
            The unitary representation.
        """
        values = values or dict()
        thetas = self.parse_values(values, embedding)
        batch_size = len(thetas)
        mat = parametric_unitary(thetas, self.operation, self.identity, batch_size)
        return mat

    def jacobian(
        self,
        values: dict[str, Tensor] | Tensor | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        """
        Get the corresponding unitary of the jacobian.

        Arguments:
            values: Parameter value.

        Returns:
            The unitary representation of the jacobian.
        """
        values = values or dict()
        thetas = self.parse_values(values, embedding)
        batch_size = len(thetas)
        return _jacobian(thetas, self.operation, self.identity, batch_size)

    def to(self, *args: Any, **kwargs: Any) -> Parametric:
        super().to(*args, **kwargs)
        self._device = self.operation.device
        self.param_name = (
            self.param_name.to(*args, **kwargs)
            if isinstance(self.param_name, torch.Tensor)
            else self.param_name
        )
        self._dtype = self.operation.dtype
        return self


class ControlledParametric(Parametric):
    """
    Primitives for controlled parametric operations.

    Attributes:
        control: Control qubit(s).
    """

    def __init__(
        self,
        operation: str | Tensor,
        control: int | Tuple[int, ...],
        target: int | Tuple[int, ...],
        param_name: str | int | float | torch.Tensor = "",
        noise: DigitalNoiseProtocol | None = None,
    ):
        """Initializes a ControlledParametric.

        Arguments:
            operation: Rotation gate.
            control: Control qubit(s).
            qubit_targets: Target qubit(s).
            param_name: Name of parameters.
        """
        support = Support(target, control)
        super().__init__(operation, support, param_name, noise=noise)

    def extra_repr(self) -> str:
        """String representation of the operation.

        Returns:
            String with information on operation.
        """
        return (
            f"control: {self.control}, target: {(self.target,)}, param: {self.param_name}"
            + _repr_noise(self.noise)
        )

    def _construct_parametric_base_op(
        self,
        values: dict[str, Tensor] | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        """
        Get the corresponding unitary with parsed values and kronned identities
        for control.

        Arguments:
            values: A dict containing a Parameter name and value.
            embedding: An optional embedding for parameters.

        Returns:
            The unitary representation.
        """
        values = values or dict()
        thetas = self.parse_values(values, embedding)
        batch_size = len(thetas)
        mat = parametric_unitary(thetas, self.operation, self.identity, batch_size)
        mat = controlled(mat, batch_size, len(self.control))
        return mat

    def jacobian(
        self,
        values: dict[str, Tensor] | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        """
        Get the corresponding unitary of the jacobian.

        Arguments:
            values: Parameter value.

        Returns:
            The unitary representation of the jacobian.
        """
        values = values or dict()
        thetas = self.parse_values(values, embedding)
        batch_size = len(thetas)
        n_control = len(self.control)
        jU = _jacobian(thetas, self.operation, self.identity, batch_size)
        n_dim = 2 ** (n_control + 1)
        jC = (
            torch.zeros((n_dim, n_dim), dtype=self.identity.dtype)
            .unsqueeze(2)
            .repeat(1, 1, batch_size)
        )
        unitary_idx = 2 ** (n_control + 1) - 2
        jC[unitary_idx:, unitary_idx:, :] = jU
        return jC


class ControlledRotationGate(ControlledParametric):
    """
    Primitives for controlled rotation operations.
    """

    n_params = 1

    def __init__(
        self,
        operation: str | Tensor,
        control: int | Tuple[int, ...],
        target: int,
        param_name: str | int | float | torch.Tensor = "",
        noise: DigitalNoiseProtocol | None = None,
    ):
        """Initializes a ControlledRotationGate.

        Arguments:
            gate: Rotation gate.
            control: Control qubit(s).
            qubit_support: Target qubit.
            param_name: Name of parameters.
        """
        super().__init__(operation, control, target, param_name, noise=noise)

    @cached_property
    def eigenvals_generator(self) -> Tensor:
        """Get eigenvalues of the underlying generator.

        Arguments:
            values: Parameter values.

        Returns:
            Eigenvalues of the generator operator.
        """
        return torch.cat(
            (
                torch.zeros(
                    2 ** len(self.qubit_support) - 2,
                    device=self.device,
                    dtype=self.dtype,
                ),
                pauli_singleq_eigenvalues.flatten().to(
                    device=self.device, dtype=self.dtype
                ),
            )
        ).reshape(-1, 1)



from __future__ import annotations

import random
from typing import Any, Callable

import pytest
import qutip
import torch
from pytest import FixtureRequest
from torch import Tensor

from pyqtorch.composite import Add, Scale, Sequence
from pyqtorch.embed import ConcretizedCallable
from pyqtorch.noise import (
    AmplitudeDamping,
    BitFlip,
    Depolarizing,
    DigitalNoiseProtocol,
    DigitalNoiseType,
    GeneralizedAmplitudeDamping,
    Noise,
    PauliChannel,
    PhaseDamping,
    PhaseFlip,
)
from pyqtorch.primitives import (
    CNOT,
    CPHASE,
    CRX,
    CRY,
    CRZ,
    CY,
    CZ,
    PHASE,
    RX,
    RY,
    RZ,
    ControlledPrimitive,
    ControlledRotationGate,
    H,
    I,
    Parametric,
    Primitive,
    S,
    T,
    X,
    Y,
    Z,
)
from pyqtorch.utils import (
    DensityMatrix,
    density_mat,
    product_state,
    random_dm_promotion,
    random_state,
)


@pytest.fixture(params=[I, X, Y, Z, H, T, S])
def gate(request: Primitive) -> Any:
    return request.param


# Parametrized fixture
@pytest.fixture
def n_qubits(request: FixtureRequest) -> Any:
    params = getattr(request, "param", {})
    low: int = params.get("low", 1)
    high: int = params.get("high", 6)
    return torch.randint(low=low, high=high, size=(1,)).item()


@pytest.fixture
def target(n_qubits: int) -> int:
    return random.choice([i for i in range(n_qubits)])


# Parametrized fixture
@pytest.fixture
def batch_size(request: FixtureRequest) -> Any:
    params = getattr(request, "param", {})
    low: int = params.get("low", 1)
    high: int = params.get("high", 5)
    return torch.randint(low=low, high=high, size=(1,)).item()


@pytest.fixture
def random_input_state(n_qubits: int, batch_size: int) -> Any:
    return random_state(n_qubits, batch_size)


@pytest.fixture
def random_input_dm(random_input_state: Tensor) -> Any:
    return density_mat(random_input_state)


@pytest.fixture
def rho_input(batch_size: int, target: int, n_qubits: int) -> Any:
    rho_0: DensityMatrix = density_mat(product_state("0", batch_size))
    rho_input: DensityMatrix = random_dm_promotion(target, rho_0, n_qubits)
    return rho_input


# TODO: create random noisy protocols
@pytest.fixture
def random_noise():
    pass


@pytest.fixture
def random_single_qubit_gate(target: int) -> Any:
    GATES = [X, Y, Z, I, H]
    gate = random.choice(GATES)
    return gate(target)


@pytest.fixture
def random_rotation_gate(target: int) -> Any:
    ROTATION_GATES = [RX, RY, RZ, PHASE]
    rotation_gate = random.choice(ROTATION_GATES)
    return rotation_gate(target, "theta")


@pytest.fixture
def random_controlled_gate(n_qubits: int, target: int) -> Any:
    if n_qubits < 2:
        raise ValueError("The controlled gates are defined on 2 qubits minimum")
    CONTROLLED_GATES = [CNOT, CY, CZ]
    controlled_gate = random.choice(CONTROLLED_GATES)
    control = random.choice([i for i in range(n_qubits) if i != target])
    return controlled_gate(control, target)


@pytest.fixture
def random_rotation_control_gate(n_qubits: int, target: int) -> Any:
    if n_qubits < 2:
        raise ValueError("The controlled gates are defined on 2 qubits minimum")
    ROTATION_CONTROL_GATES = [CRX, CRY, CRZ, CPHASE]
    controlled_gate = random.choice(ROTATION_CONTROL_GATES)
    control = random.choice([i for i in range(n_qubits) if i != target])
    return controlled_gate(control, target, "theta")


@pytest.fixture
def random_unitary_gate(
    random_single_qubit_gate: Primitive,
    random_rotation_gate: Parametric,
    random_controlled_gate: ControlledPrimitive,
    random_rotation_control_gate: ControlledRotationGate,
) -> Any:
    UNITARY_GATES = [
        random_single_qubit_gate,
        random_controlled_gate,
        random_rotation_gate,
        random_rotation_control_gate,
    ]
    unitary_gate = random.choice(UNITARY_GATES)
    return unitary_gate


@pytest.fixture
def random_noise_gate(n_qubits: int) -> Any:
    NOISE_GATES = [
        BitFlip,
        PhaseFlip,
        PauliChannel,
        Depolarizing,
        AmplitudeDamping,
        PhaseDamping,
        GeneralizedAmplitudeDamping,
    ]
    noise_gate = random.choice(NOISE_GATES)
    if noise_gate == PauliChannel:
        noise_prob = torch.rand(size=(3,))
        noise_prob = noise_prob / (
            noise_prob.sum(dim=0, keepdim=True) + torch.rand((1,)).item()
        )
        return noise_gate(torch.randint(0, n_qubits, (1,)).item(), noise_prob)
    elif noise_gate == GeneralizedAmplitudeDamping:
        noise_prob = torch.rand(size=(2,))
        return noise_gate(torch.randint(0, n_qubits, (1,)).item(), noise_prob)
    else:
        noise_prob = torch.rand(size=(1,)).item()
        return noise_gate(torch.randint(0, n_qubits, (1,)).item(), noise_prob)


@pytest.fixture
def random_noisy_protocol() -> Any:
    noise_type = random.choice(list(DigitalNoiseType))
    if noise_type == DigitalNoiseType.PAULI_CHANNEL:
        noise_prob = torch.rand(size=(3,))
        noise_prob = noise_prob / (
            noise_prob.sum(dim=0, keepdim=True) + torch.rand((1,)).item()
        )
    elif noise_type == DigitalNoiseType.GENERALIZED_AMPLITUDE_DAMPING:
        noise_prob = torch.rand(size=(2,))
    else:
        noise_prob = torch.rand(size=(1,)).item()
    return DigitalNoiseProtocol(noise_type, noise_prob)


@pytest.fixture
def random_noisy_unitary_gate(
    n_qubits: int, target: int, random_noisy_protocol: DigitalNoiseProtocol
) -> Any:
    if n_qubits < 2:
        raise ValueError("The controlled gates are defined on 2 qubits minimum")
    SINGLE_GATES = [X, Y, Z, I, H]
    ROTATION_GATES = [RX, RY, RZ, PHASE]
    CONTROLLED_GATES = [CNOT, CY, CZ]
    ROTATION_CONTROL_GATES = [CRX, CRY, CRZ, CPHASE]
    UNITARY_GATES = (
        SINGLE_GATES + ROTATION_GATES + CONTROLLED_GATES + ROTATION_CONTROL_GATES
    )
    unitary_gate = random.choice(UNITARY_GATES)
    protocol_gates, protocol_info = random_noisy_protocol.gates[0]
    noise_gate = protocol_gates(target, protocol_info.error_probability)
    if unitary_gate in SINGLE_GATES:
        return (
            unitary_gate(target=target, noise=random_noisy_protocol),  # type: ignore[call-arg]
            unitary_gate(target=target),  # type: ignore[call-arg]
            noise_gate,
        )  # type: ignore[call-arg]
    if unitary_gate in ROTATION_GATES:
        return (
            unitary_gate(
                target=target, param_name="theta", noise=random_noisy_protocol
            ),  # type: ignore[call-arg]
            unitary_gate(target=target, param_name="theta"),  # type: ignore[call-arg]
            noise_gate,
        )
    if unitary_gate in CONTROLLED_GATES:
        control = random.choice([i for i in range(n_qubits) if i != target])
        return (
            unitary_gate(control=control, target=target, noise=random_noisy_protocol),  # type: ignore[call-arg]
            unitary_gate(control=control, target=target),  # type: ignore[call-arg]
            noise_gate,
        )
    if unitary_gate in ROTATION_CONTROL_GATES:
        control = random.choice([i for i in range(n_qubits) if i != target])
        return (
            unitary_gate(
                control=control,
                target=target,
                param_name="theta",
                noise=random_noisy_protocol,
            ),  # type: ignore[call-arg]
            unitary_gate(control=control, target=target, param_name="theta"),  # type: ignore[call-arg]
            noise_gate,
        )


@pytest.fixture
def random_flip_gate() -> Any:
    FLIP_GATES = [BitFlip, PhaseFlip, Depolarizing, PauliChannel]
    # The gate is not initialize
    return random.choice(FLIP_GATES)


@pytest.fixture
def flip_probability(random_flip_gate: Noise) -> Any:
    if random_flip_gate == PauliChannel:
        probabilities = torch.rand((3))
        probabilities = probabilities / (
            torch.sum(probabilities) + torch.rand((1,)).item()
        )  # To ensure that the sum of the probabilities is lower than 1.
    else:
        probabilities = torch.rand(1).item()
    return probabilities


@pytest.fixture
def flip_expected_state(
    n_qubits: int,
    target: int,
    random_flip_gate: Noise,
    flip_probability: Tensor,
    batch_size: int,
) -> Any:
    if random_flip_gate == BitFlip:
        expected_state = DensityMatrix(
            torch.tensor(
                [[[1 - flip_probability], [0]], [[0], [flip_probability]]],
                dtype=torch.cdouble,
            )
        )
    elif random_flip_gate == PhaseFlip:
        expected_state = DensityMatrix(
            torch.tensor([[[1], [0]], [[0], [0]]], dtype=torch.cdouble)
        )
    elif random_flip_gate == Depolarizing:
        expected_state = DensityMatrix(
            torch.tensor(
                [
                    [[1 - (2 * flip_probability) / 3], [0]],
                    [[0], [(2 * flip_probability) / 3]],
                ],
                dtype=torch.cdouble,
            )
        )
    elif random_flip_gate == PauliChannel:
        px, py = flip_probability[0], flip_probability[1]
        expected_state = DensityMatrix(
            torch.tensor(
                [[[1 - (px + py)], [0]], [[0], [px + py]]], dtype=torch.cdouble
            )
        )
    expected_state = random_dm_promotion(target, expected_state, n_qubits)
    return expected_state.repeat(1, 1, batch_size)


@pytest.fixture
def flip_gates_prob_0(random_flip_gate: Noise, target: int) -> Any:
    if random_flip_gate == PauliChannel:
        FlipGate_0 = random_flip_gate(target, error_probability=(0, 0, 0))
    else:
        FlipGate_0 = random_flip_gate(target, error_probability=0)
    return FlipGate_0


@pytest.fixture
def flip_gates_prob_1(
    random_flip_gate: Noise, target: int, random_input_state: Tensor
) -> Any:
    if random_flip_gate == BitFlip:
        FlipGate_1 = random_flip_gate(target, error_probability=1)
        expected_op = density_mat(X(target)(random_input_state))
    elif random_flip_gate == PhaseFlip:
        FlipGate_1 = random_flip_gate(target, error_probability=1)
        expected_op = density_mat(Z(target)(random_input_state))
    elif random_flip_gate == Depolarizing:
        FlipGate_1 = random_flip_gate(target, error_probability=1)
        expected_op = (
            1
            / 3
            * (
                density_mat(Z(target)(random_input_state))
                + density_mat(X(target)(random_input_state))
                + density_mat(Y(target)(random_input_state))
            )
        )
    elif random_flip_gate == PauliChannel:
        px, py, pz = 1 / 3, 1 / 3, 1 / 3
        FlipGate_1 = random_flip_gate(target, error_probability=(px, py, pz))
        expected_op = (
            px * density_mat(X(target)(random_input_state))  # type: ignore
            + py * density_mat(Y(target)(random_input_state))
            + pz * density_mat(Z(target)(random_input_state))
        )
    return FlipGate_1, expected_op


@pytest.fixture
def random_damping_gate() -> Any:
    DAMPING_GATES = [AmplitudeDamping, PhaseDamping, GeneralizedAmplitudeDamping]
    # The gate is not initialize
    return random.choice(DAMPING_GATES)


@pytest.fixture
def damping_rate(random_damping_gate: Noise) -> Any:
    if random_damping_gate == GeneralizedAmplitudeDamping:
        rate = torch.rand((2,))  # prob and rate
    else:
        rate = torch.rand(1).item()
    return rate


@pytest.fixture
def damping_expected_state(
    n_qubits: int,
    target: int,
    random_damping_gate: Noise,
    damping_rate: Tensor,
    batch_size: int,
    rho_input: Tensor,
) -> Any:
    if random_damping_gate == GeneralizedAmplitudeDamping:
        p, r = damping_rate[0], damping_rate[1]
        expected_state = DensityMatrix(
            torch.tensor(
                [[[1 - (r - p * r)], [0]], [[0], [r - p * r]]], dtype=torch.cdouble
            )
        )
        DampingGate = random_damping_gate(target, damping_rate)
        expected_state = random_dm_promotion(target, expected_state, n_qubits).repeat(
            1, 1, batch_size
        )
    else:
        expected_state = I(target)(rho_input)
        DampingGate = random_damping_gate(target, damping_rate)
    return DampingGate, expected_state


@pytest.fixture
def damping_gates_prob_0(random_damping_gate: Noise, target: int) -> Any:
    if random_damping_gate == GeneralizedAmplitudeDamping:
        damping_gate_0 = random_damping_gate(target, (0, 0))
    else:
        damping_gate_0 = random_damping_gate(target, 0)
    return damping_gate_0


@pytest.fixture
def duration() -> float:
    return float(torch.rand(1))


@pytest.fixture
def n_steps() -> float:
    return int(torch.randint(100, 1000, (1,)))


@pytest.fixture
def omega() -> float:
    return 20.0


@pytest.fixture
def param_x() -> float:
    return float(5.0 * torch.rand(1))


@pytest.fixture
def param_y() -> float:
    return float(2.0 * torch.rand(1))


@pytest.fixture
def sigma_x() -> Tensor:
    return torch.tensor([[0.0, 1.0], [1.0, 0.0]])


@pytest.fixture
def sigma_y() -> Tensor:
    return torch.tensor([[0.0, -1.0j], [1.0j, 0.0]])


@pytest.fixture
def jump_op_torch() -> Tensor:
    return [torch.eye(4, dtype=torch.complex128)]


@pytest.fixture
def jump_op_qutip() -> Tensor:
    return [qutip.qeye(4)]


@pytest.fixture
def torch_hamiltonian(
    omega: float, param_x: float, param_y: float, sigma_x: Tensor, sigma_y: Tensor
) -> Callable:
    def hamiltonian_t(t: float) -> Tensor:
        t = torch.as_tensor(t)
        return omega * (
            param_y * torch.sin(t) * torch.kron(sigma_x, torch.eye(2))
            + param_x * t**2 * torch.kron(torch.eye(2), sigma_y)
        ).to(torch.complex128)

    return hamiltonian_t


@pytest.fixture
def qutip_hamiltonian(omega: float, param_x: float, param_y: float) -> Callable:
    def hamiltonian_t(t: float, args: Any) -> qutip.Qobj:
        return qutip.Qobj(
            omega
            * (
                param_y
                * torch.sin(torch.as_tensor(t)).numpy()
                * qutip.tensor(qutip.sigmax(), qutip.qeye(2))
                + param_x * t**2 * qutip.tensor(qutip.qeye(2), qutip.sigmay())
            ).full()
        )

    return hamiltonian_t


@pytest.fixture
def tparam() -> str:
    return "t"


@pytest.fixture
def sin(tparam: str) -> tuple[str, ConcretizedCallable]:
    sin_t, sin_fn = "sin_x", ConcretizedCallable("sin", [tparam])
    return sin_t, sin_fn


@pytest.fixture
def sq(tparam: str) -> tuple[str, ConcretizedCallable]:
    t_sq, t_sq_fn = "t_sq", ConcretizedCallable("mul", [tparam, tparam])
    return t_sq, t_sq_fn


@pytest.fixture
def hamevo_generator(
    omega: float, param_x: float, param_y: float, sin: tuple, sq: tuple
) -> Sequence:
    sin_t, _ = sin
    t_sq, _ = sq
    generator = Scale(
        Add(
            [
                Scale(Scale(X(0), sin_t), "y"),
                Scale(Scale(Y(1), t_sq), param_x),
            ]
        ),
        omega,
    )
    return generator



from __future__ import annotations

from typing import Any, Callable

import torch
from torch import Tensor

from pyqtorch.time_dependent.options import AdaptiveSolverOptions
from pyqtorch.time_dependent.solvers import MEDormandPrince5
from pyqtorch.utils import Result, SolverType


def mesolve(
    H: Callable[..., Any],
    rho0: Tensor,
    L: list[Tensor],
    tsave: list | Tensor,
    solver: SolverType,
    options: dict[str, Any] | None = None,
) -> Result:
    """Solve time-dependent Lindblad master equation.

    Args:
        H (Callable[[float], Tensor]): time-dependent Hamiltonian of the system
        rho0 (Tensor): initial density matrix of the system
        L (list[Tensor]): list of jump operators
        tsave (Tensor): tensor containing simulation time instants
        solver (SolverType): name of the solver to use
        options (dict[str, Any], optional): additional options passed to the solver.
            Defaults to None.

    Returns:
        Result: dataclass containing the simulated density matrices at each time moment
    """
    options = options or dict()
    L = torch.stack(L)

    # check dimensions of initial state
    n = H(0.0).shape[0]
    if (
        (rho0.shape[0] != rho0.shape[1])
        or (rho0.shape[0] != n)
        or (len(rho0.shape) != 3)
    ):
        raise ValueError(
            f"Argument `rho0` must be a 3D tensor of shape `({n}, {n}, batch_size)`. "
            f"Current shape: {tuple(rho0.shape)}."
        )

    # permute dimensions to allow batch operations
    rho0 = rho0.permute(2, 0, 1)

    # instantiate appropriate solver
    if solver == SolverType.DP5_ME:
        opt = AdaptiveSolverOptions(**options)
        s = MEDormandPrince5(H, rho0, L, tsave, opt)
    else:
        raise ValueError("Requested solver is not available.")

    # compute the result
    result = s.run()

    return Result(result)



from __future__ import annotations

import logging
from functools import cached_property
from logging import getLogger
from math import log2
from typing import Any, Callable

import torch
from torch import Tensor

from pyqtorch.apply import apply_operator, apply_operator_dm
from pyqtorch.embed import Embedding
from pyqtorch.matrices import _dagger
from pyqtorch.noise import DigitalNoiseProtocol, _repr_noise
from pyqtorch.qubit_support import Support
from pyqtorch.utils import (
    DensityMatrix,
    density_mat,
    expand_operator,
    is_diag_batched,
    permute_basis,
)

logger = getLogger(__name__)


def forward_hook(*args, **kwargs) -> None:  # type: ignore[no-untyped-def]
    torch.cuda.nvtx.range_pop()


def pre_forward_hook(*args, **kwargs) -> None:  # type: ignore[no-untyped-def]
    torch.cuda.nvtx.range_push("QuantumOperation.forward")


def backward_hook(*args, **kwargs) -> None:  # type: ignore[no-untyped-def]
    torch.cuda.nvtx.range_pop()


def pre_backward_hook(*args, **kwargs) -> None:  # type: ignore[no-untyped-def]
    torch.cuda.nvtx.range_push("QuantumOperation.backward")


class QuantumOperation(torch.nn.Module):
    """Generic QuantumOperation class storing a tensor operation to represent either
        a quantum operator or a tensor generator inferring the QuantumOperation.

    Attributes:
        operation (Tensor): Tensor used to infer the QuantumOperation
            directly or indirectly.
        qubit_support (int | tuple[int, ...]): Tuple of qubits
            the QuantumOperation acts on.
        operator_function (Callable | None, optional): Function to generate the base operator
            from operation. If None, we consider returning operation itself.
        noise ( NoiseProtocol | dict[str, NoiseProtocol. optional): Type of noise
            to add in the operation.
        diagonal (bool, optional): Specify if the operation is diagonal.
            For supporting, only pass a 2-dim operation tensor
            containing diagonal elements with batchsize.

    """

    def __init__(
        self,
        operation: Tensor,
        qubit_support: int | tuple[int, ...] | Support,
        operator_function: Callable | None = None,
        noise: DigitalNoiseProtocol | None = None,
        diagonal: bool = False,
    ) -> None:
        """Initializes QuantumOperation

        Args:
            operation (Tensor): Tensor used to infer the QuantumOperation.
            qubit_support (int | tuple[int, ...]): List of qubits
                the QuantumOperation acts on.

        Raises:
            ValueError: When operation has incompatible shape
                with  qubit_support
        """
        super().__init__()

        self._qubit_support = (
            qubit_support
            if isinstance(qubit_support, Support)
            else Support(target=qubit_support)
        )
        # to inform on diagonality of operation
        self._diagonal = diagonal
        self.is_diagonal: bool = False
        if diagonal:
            if len(operation.size()) != 2:
                raise ValueError(
                    "Only pass diagonal operators as 2D tensors, with batchsize."
                )
            self.is_diagonal = diagonal
        else:
            self.is_diagonal = is_diag_batched(operation)

        self.register_buffer("operation", operation)
        self._device = self.operation.device
        self._dtype = self.operation.dtype

        is_primitive = operator_function is None
        dim_nomatch = len(self.qubit_support) != int(log2(operation.shape[0]))
        if is_primitive and dim_nomatch:
            raise ValueError(
                "The operation shape should match the length of the qubit_support."
            )

        if operator_function is None:
            self._operator_function = self._default_operator_function
        else:
            self._operator_function = operator_function

        self.noise = noise

        if logger.isEnabledFor(logging.DEBUG):
            # When Debugging let's add logging and NVTX markers
            # WARNING: incurs performance penalty
            self.register_forward_hook(forward_hook, always_call=True)
            self.register_full_backward_hook(backward_hook)
            self.register_forward_pre_hook(pre_forward_hook)
            self.register_full_backward_pre_hook(pre_backward_hook)

    def to(self, *args: Any, **kwargs: Any) -> QuantumOperation:
        """Do device or dtype conversions.

        Returns:
            QuantumOperation: Converted instance.
        """
        super().to(*args, **kwargs)
        self._device = self.operation.device
        self._dtype = self.operation.dtype
        return self

    @property
    def qubit_support(self) -> tuple[int, ...]:
        """Getter qubit_support.

        Returns:
            Support: Tuple of sorted qubits.
        """
        return self._qubit_support.sorted_qubits

    @property
    def target(self) -> tuple[int, ...]:
        """Get target qubits.

        Returns:
            tuple[int, ...]: The target qubits
        """
        return self._qubit_support.target

    @property
    def control(self) -> tuple[int, ...]:
        """Get control qubits.

        Returns:
            tuple[int, ...]: The control qubits
        """
        return self._qubit_support.control

    @property
    def operator_function(self) -> Callable[..., Any]:
        """Getter operator_function.

        Returns:
            Callable[..., Any]: Function for
                getting base operator.
        """
        return self._operator_function

    def __hash__(self) -> int:
        return hash(self.qubit_support)

    def extra_repr(self) -> str:
        return f"target: {self.qubit_support}" + _repr_noise(self.noise)

    @property
    def is_parametric(self) -> bool:
        """Consider by default that operation is parametric when `operator_function` is not None.

        Returns:
            bool: True if `operator_function` is not None
        """
        return self.operator_function is not None

    @property
    def device(self) -> torch.device:
        """Returns device.

        Returns:
            torch.device: Device.
        """
        return self._device

    @property
    def dtype(self) -> torch.dtype:
        """Returns dtype.

        Returns:
            torch.dtype: Dtype.
        """
        return self._dtype

    @cached_property
    def eigenvals_generator(self) -> Tensor:
        """Get eigenvalues of the underlying operation.

        Arguments:
            values: Parameter values.

        Returns:
            Eigenvalues of the operation.
        """
        return torch.linalg.eigvalsh(self.operation).reshape(-1, 1)

    @cached_property
    def eigenvalues(
        self,
        values: dict[str, Tensor] | Tensor | None = None,
        embedding: Embedding | None = None,
        diagonal: bool = False,
    ) -> Tensor:
        """Get eigenvalues of the tensor of QuantumOperation.

        Args:
            values (dict[str, Tensor], optional): Parameter values. Defaults to dict().
            embedding (Embedding | None, optional): Optional embedding. Defaults to None.

        Returns:
            Eigenvalues of the related tensor.
        """
        blockmat = self.tensor(values or dict(), embedding)
        if len(blockmat.shape) == 3:
            return torch.linalg.eigvals(blockmat.permute((2, 0, 1))).reshape(-1, 1)
        else:
            # for diagonal cases
            return blockmat

    @cached_property
    def spectral_gap(self) -> Tensor:
        """Difference between the moduli of the two largest eigenvalues of the generator.

        Returns:
            Tensor: Spectral gap value.
        """
        spectrum = self.eigenvals_generator
        diffs = spectrum - spectrum.T
        spectral_gap = torch.unique(torch.abs(torch.tril(diffs)))
        return spectral_gap[spectral_gap.nonzero()]

    def _default_operator_function(
        self,
        values: dict[str, Tensor] | Tensor | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        """Default operator_function simply returns the operation.

        Args:
            values (dict[str, Tensor] | Tensor, optional): Parameter values. Defaults to dict().
            embedding (Embedding | None, optional): Optional embedding. Defaults to None.

        Returns:
            Tensor: Base operator.
        """
        operation = (
            self.operation.unsqueeze(2)
            if (len(self.operation.shape) == 2 and not self._diagonal)
            else self.operation
        )
        return operation

    def dagger(
        self,
        values: dict[str, Tensor] | Tensor | None = None,
        embedding: Embedding | None = None,
        diagonal: bool = False,
    ) -> Tensor:
        """Apply the dagger to operator.

        Args:
            values (dict[str, Tensor] | Tensor, optional): Parameter values. Defaults to dict().
            embedding (Embedding | None, optional): Optional embedding. Defaults to None.
            diagonal (bool, optional): Whether to return the diagonal form of the tensor or not.
                Defaults to False.

        Returns:
            Tensor: conjugate transpose operator.
        """
        blockmat = self.operator_function(values or dict(), embedding)
        use_diagonal = diagonal and self.is_diagonal
        return _dagger(blockmat, use_diagonal)

    def tensor(
        self,
        values: dict[str, Tensor] | None = None,
        embedding: Embedding | None = None,
        full_support: tuple[int, ...] | None = None,
        diagonal: bool = False,
    ) -> Tensor:
        """Get tensor of the QuantumOperation.

        Args:
            values (dict[str, Tensor], optional): Parameter values. Defaults to dict().
            embedding (Embedding | None, optional): Optional embedding. Defaults to None.
            full_support (tuple[int, ...] | None, optional): The qubits the returned tensor
                will be defined over. Defaults to None for only using the qubit_support.
            diagonal (bool, optional): Whether to return the diagonal form of the tensor or not.
                Defaults to False.

        Returns:
            Tensor: Tensor representation of QuantumOperation.
        """
        values = values or dict()
        blockmat = self.operator_function(values, embedding)
        use_diagonal = diagonal and self.is_diagonal
        if use_diagonal and not self._diagonal:
            blockmat = torch.diagonal(blockmat).T
        if self._qubit_support.qubits != self.qubit_support:
            blockmat = permute_basis(
                blockmat, self._qubit_support.qubits, inv=True, diagonal=use_diagonal
            )
        if full_support is None:
            return blockmat
        else:
            return expand_operator(
                blockmat, self.qubit_support, full_support, diagonal=use_diagonal
            )

    def _forward(
        self,
        state: Tensor,
        values: dict[str, Tensor] | Tensor | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        values = values or dict()
        if isinstance(state, DensityMatrix):
            return apply_operator_dm(
                state, self.tensor(values, embedding), self.qubit_support
            )
        else:
            return apply_operator(
                state,
                self.tensor(values, embedding),
                self.qubit_support,
            )

    def _noise_forward(
        self,
        state: Tensor,
        values: dict[str, Tensor] | Tensor | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        values = values or dict()
        if not isinstance(state, DensityMatrix):
            state = density_mat(state)

        state = apply_operator_dm(
            state, self.tensor(values, embedding), self.qubit_support
        )

        for noise_class, noise_info in self.noise.gates:  # type: ignore [union-attr]
            if noise_info.target is None:
                target = self.target if len(self.target) == 1 else self.target[0]
            else:
                target = noise_info.target
            noise_gate = noise_class(
                target=target, error_probability=noise_info.error_probability
            )
            state = noise_gate(state, values)

        return state

    def forward(
        self,
        state: Tensor,
        values: dict[str, Tensor] | Tensor | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        """Apply the operation on input state or density matrix.

        Args:
            state (Tensor): Input state.
            values (dict[str, Tensor], optional): Parameter values. Defaults to dict().
            embedding (Embedding | None, optional): Optional embedding. Defaults to None.

        Returns:
            Tensor: _description_
        """
        values = values or dict()
        if self.noise:
            return self._noise_forward(state, values, embedding)
        else:
            return self._forward(state, values, embedding)



from __future__ import annotations

import random
from math import log2
from typing import Callable, Tuple

import pytest
import torch
from torch import Tensor

import pyqtorch as pyq
from pyqtorch import ConcretizedCallable
from pyqtorch.apply import apply_operator
from pyqtorch.matrices import (
    DEFAULT_MATRIX_DTYPE,
)
from pyqtorch.primitives import Parametric, Primitive
from pyqtorch.utils import (
    ATOL,
    product_state,
)

state_000 = product_state("000")
state_001 = product_state("001")
state_100 = product_state("100")
state_101 = product_state("101")
state_110 = product_state("110")
state_011 = product_state("011")
state_111 = product_state("111")
state_0000 = product_state("0000")
state_1110 = product_state("1110")
state_1111 = product_state("1111")

device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")


def test_identity() -> None:
    assert torch.allclose(product_state("0"), pyq.I(0)(product_state("0"), None))
    assert torch.allclose(product_state("1"), pyq.I(0)(product_state("1")))


def test_N() -> None:
    null_state = torch.zeros_like(pyq.zero_state(1))
    assert torch.allclose(null_state, pyq.N(0)(product_state("0"), None))
    assert torch.allclose(product_state("1"), pyq.N(0)(product_state("1"), None))


@pytest.mark.parametrize("n_qubits", [1, 2, 3])
def test_projectors(n_qubits: int) -> None:
    qubit_support = tuple(range(n_qubits))
    max_possibility = 2**n_qubits
    possible_nbs = list(range(max_possibility - 1))
    for ket in [random.choice(possible_nbs) for _ in range(2 * n_qubits)]:
        for bra in [random.choice(possible_nbs) for _ in range(2 * n_qubits)]:
            t_mat = torch.zeros(max_possibility, dtype=DEFAULT_MATRIX_DTYPE)
            t_mat[ket] = 1.0 + 0.0j
            t_mat = t_mat.reshape((2,) * n_qubits + (1,))
            assert torch.allclose(
                pyq.Projector(
                    qubit_support, ket=f"{ket:0{n_qubits}b}", bra=f"{bra:0{n_qubits}b}"
                )(product_state(f"{bra:0{n_qubits}b}")),
                t_mat,
            )
        if ket > 0:
            assert torch.allclose(
                pyq.Projector(
                    qubit_support, ket=f"{ket:0{n_qubits}b}", bra=f"{ket:0{n_qubits}b}"
                )(product_state(f"{ket-1:0{n_qubits}b}")),
                torch.zeros((2,) * n_qubits + (1,), dtype=DEFAULT_MATRIX_DTYPE),
            )


czop_example = pyq.CZ(control=(0, 1), target=2)
crxop_example = pyq.CRX(control=(0, 1), target=2, param_name="theta")


@pytest.mark.parametrize(
    "projector, initial_state, final_state",
    [
        (czop_example, state_101, state_101),
        (czop_example, state_111, -state_111),
        (crxop_example, state_001, state_001),
        (
            crxop_example,
            state_110,
            torch.tensor(
                [
                    [
                        [[0.0000 + 0.0000j], [0.0000 + 0.0000j]],
                        [[0.0000 + 0.0000j], [0.0000 + 0.0000j]],
                    ],
                    [
                        [[0.0000 + 0.0000j], [0.0000 + 0.0000j]],
                        [[0.8776 + 0.0000j], [0.0000 - 0.4794j]],
                    ],
                ],
                dtype=torch.complex128,
            ),
        ),
        (
            crxop_example,
            state_111,
            torch.tensor(
                [
                    [
                        [[0.0000 + 0.0000j], [0.0000 + 0.0000j]],
                        [[0.0000 + 0.0000j], [0.0000 + 0.0000j]],
                    ],
                    [
                        [[0.0000 + 0.0000j], [0.0000 + 0.0000j]],
                        [[0.0000 - 0.4794j], [0.8776 + 0.0000j]],
                    ],
                ],
                dtype=torch.complex128,
            ),
        ),
    ],
)
def test_multicontrol_rotation(
    projector: Primitive, initial_state: torch.Tensor, final_state: torch.Tensor
) -> None:

    val_param = {"theta": torch.Tensor([1.0])}
    projector_apply_res = projector(initial_state, val_param)
    assert torch.allclose(final_state, projector_apply_res, atol=1.0e-4)


def test_CNOT_state00_controlqubit_0() -> None:
    result: Tensor = pyq.CNOT(0, 1)(product_state("00"), None)
    assert torch.equal(product_state("00"), result)


def test_CNOT_state10_controlqubit_0() -> None:
    result: Tensor = pyq.CNOT(0, 1)(product_state("10"), None)
    assert torch.allclose(product_state("11"), result)


def test_CNOT_state11_controlqubit_0() -> None:
    result: Tensor = pyq.CNOT(0, 1)(product_state("11"), None)
    assert torch.allclose(product_state("10"), result)


def test_CRY_state10_controlqubit_0() -> None:
    result: Tensor = pyq.CRY(0, 1, "theta")(
        product_state("10"), {"theta": torch.tensor([torch.pi])}
    )
    assert torch.allclose(product_state("11"), result, atol=ATOL)


def test_CRY_state01_controlqubit_0() -> None:
    result: Tensor = pyq.CRY(1, 0, "theta")(
        product_state("01"), {"theta": torch.tensor([torch.pi])}
    )
    assert torch.allclose(product_state("11"), result, atol=ATOL)


@pytest.mark.parametrize(
    "initial_state,expected_state",
    [
        (state_000, state_000),
        (state_001, state_001),
        (state_100, state_100),
        (state_101, state_110),
        (state_110, state_101),
    ],
)
def test_CSWAP_controlqubits0(initial_state: Tensor, expected_state: Tensor) -> None:
    cswap = pyq.CSWAP(0, (1, 2))
    assert torch.allclose(cswap(initial_state, None), expected_state)


@pytest.mark.parametrize(
    "initial_state,expected_state",
    [
        (state_000, state_000),
        (state_001, state_001),
        (state_100, state_100),
        (state_101, state_101),
        (state_110, state_111),
        (state_1110, state_1111),
    ],
)
def test_Toffoli_controlqubits0(initial_state: Tensor, expected_state: Tensor) -> None:
    n_qubits = int(log2(torch.numel(initial_state)))
    qubits = tuple([i for i in range(n_qubits)])
    toffoli = pyq.Toffoli(qubits[:-1], qubits[-1])
    assert torch.allclose(toffoli(initial_state, None), expected_state)


@pytest.mark.parametrize(
    "initial_state,expects_rotation",
    [
        (state_000, False),
        (state_001, False),
        (state_100, False),
        (state_101, False),
        (state_110, True),
        (state_1110, True),
    ],
)
@pytest.mark.parametrize("gate", ["RX", "RY", "RZ", "PHASE"])
@pytest.mark.parametrize("batch_size", [1, 2])
@pytest.mark.parametrize("dtype", [torch.complex64, torch.complex128])
def test_multi_controlled_gates(
    initial_state: Tensor,
    expects_rotation: bool,
    batch_size: int,
    gate: str,
    dtype: torch.dtype,
) -> None:
    phi = "phi"

    initial_state = initial_state.to(device=device, dtype=dtype)
    rot_gate = getattr(pyq, gate)
    controlled_rot_gate = getattr(pyq, "C" + gate)
    phi = torch.rand(batch_size).to(device=device)
    n_qubits = int(log2(torch.numel(initial_state)))
    qubits = tuple([i for i in range(n_qubits)])
    op = controlled_rot_gate(qubits[:-1], qubits[-1], "phi").to(
        device=device, dtype=dtype
    )
    out = op(initial_state, {"phi": phi})
    expected_state = (
        rot_gate(qubits[-1], "phi").to(device=device, dtype=dtype)(
            initial_state, {"phi": phi}
        )
        if expects_rotation
        else initial_state
    )
    assert torch.allclose(out, expected_state)
    if gate != "PHASE":
        assert len(op.spectral_gap) == 2
    else:
        assert op.spectral_gap == 2.0

    assert op.eigenvals_generator.dtype == dtype


@pytest.mark.parametrize(
    "state_fn", [pyq.random_state, pyq.zero_state, pyq.uniform_state]
)
@pytest.mark.parametrize("batch_size", [1, 2, 4])
@pytest.mark.parametrize("n_qubits", [1, 2, 4])
def test_parametrized_phase_gate(
    state_fn: Callable, batch_size: int, n_qubits: int
) -> None:
    target: int = torch.randint(low=0, high=n_qubits, size=(1,)).item()
    state = state_fn(n_qubits, batch_size=batch_size)
    phi = torch.tensor([torch.pi / 2], dtype=DEFAULT_MATRIX_DTYPE)
    phase = pyq.PHASE(target, "phi")
    constant_phase = pyq.S(target)
    assert torch.allclose(phase(state, {"phi": phi}), constant_phase(state, None))
    assert phase.spectral_gap == 2.0
    assert constant_phase.spectral_gap == 1.0


def test_dagger_single_qubit() -> None:
    for cls in [
        pyq.X,
        pyq.Y,
        pyq.Z,
        pyq.S,
        pyq.H,
        pyq.T,
        pyq.RX,
        pyq.RY,
        pyq.RZ,
        pyq.PHASE,
    ]:
        n_qubits = torch.randint(low=1, high=4, size=(1,)).item()
        target = random.choice([i for i in range(n_qubits)])
        state = pyq.random_state(n_qubits)
        for param_name in ["theta", ""]:
            if issubclass(cls, Parametric):
                op = cls(target, param_name)  # type: ignore[arg-type]
            else:
                op = cls(target)  # type: ignore[assignment, call-arg]
            values = (
                {param_name: torch.rand(1)} if param_name == "theta" else torch.rand(1)
            )
            new_state = apply_operator(state, op.tensor(values), [target])
            daggered_back = apply_operator(new_state, op.dagger(values), [target])
            assert torch.allclose(daggered_back, state)


def test_dagger_nqubit() -> None:
    for cls in [
        pyq.SWAP,
        pyq.CNOT,
        pyq.CY,
        pyq.CZ,
        pyq.CRX,
        pyq.CRY,
        pyq.CRZ,
        pyq.CPHASE,
    ]:
        qubit_support: Tuple[int, ...]
        n_qubits = torch.randint(low=3, high=8, size=(1,)).item()
        target = random.choice([i for i in range(n_qubits - 2)])
        state = pyq.random_state(n_qubits)
        for param_name in ["theta", ""]:
            if isinstance(cls, (pyq.CSWAP, pyq.Toffoli)):
                op = cls((target - 2, target - 1), target)
                qubit_support = (target + 2, target + 1, target)
            elif issubclass(cls, Parametric):
                op = cls(target - 1, target, param_name)  # type: ignore[arg-type]
                qubit_support = (target + 1, target)
            else:
                op = cls(target - 1, target)  # type: ignore[call-arg]
                qubit_support = (target + 1, target)
            values = (
                {param_name: torch.rand(1)} if param_name == "theta" else torch.rand(1)
            )
            new_state = apply_operator(state, op.tensor(values), qubit_support)
            daggered_back = apply_operator(new_state, op.dagger(values), qubit_support)
            assert torch.allclose(daggered_back, state)


@pytest.mark.parametrize("dtype", [torch.complex64, torch.complex128])
def test_U(dtype: torch.dtype) -> None:
    n_qubits = torch.randint(low=1, high=8, size=(1,)).item()
    target = random.choice([i for i in range(n_qubits)])
    params = ["phi", "theta", "omega"]
    u = pyq.U(target, phi=params[0], theta=params[1], omega=params[2]).to(dtype=dtype)
    values = {param: torch.rand(1) for param in params}
    state = pyq.random_state(n_qubits, dtype=dtype)
    fwd_u = u(state, values)
    circ_u = pyq.QuantumCircuit(n_qubits, u.digital_decomposition()).to(dtype=dtype)
    fwd_circ_u = circ_u(state, values)
    assert fwd_u.dtype == fwd_circ_u.dtype == dtype
    assert torch.allclose(
        fwd_u,
        fwd_circ_u,
    )
    assert u.spectral_gap == 2.0


@pytest.mark.parametrize("gate", [pyq.RX, pyq.RY, pyq.RZ])
def test_parametric_constantparam(gate: Parametric) -> None:
    n_qubits = 4
    max_batch_size = 10
    target = torch.randint(0, n_qubits, (1,)).item()
    param_val = torch.rand(torch.randint(1, max_batch_size, (1,)).item())
    state = pyq.random_state(n_qubits)
    assert torch.allclose(
        gate(target, "theta")(state, {"theta": param_val}),
        gate(target, param_val)(state),
    )


@pytest.mark.parametrize("gate", [pyq.RX, pyq.RY, pyq.RZ])
def test_parametric_callableparam(gate: Parametric) -> None:
    n_qubits = 4
    max_batch_size = 10
    target = torch.randint(0, n_qubits, (1,)).item()
    size = torch.randint(1, max_batch_size, (1,)).item()
    param_val_x = torch.rand(size)
    param_val_y = torch.rand(size)
    state = pyq.random_state(n_qubits)
    param = ConcretizedCallable("add", ["x", "y"])
    assert torch.allclose(
        gate(target, param)(state, {"x": param_val_x, "y": param_val_y}),
        gate(target, param_val_x + param_val_y)(state),
    )



from __future__ import annotations

from typing import Callable

import pytest
import torch
from helpers import random_parameter_names, random_pauli_hamiltonian

import pyqtorch as pyq
from pyqtorch import DiffMode, expectation
from pyqtorch.circuit import QuantumCircuit
from pyqtorch.hamiltonians import HamiltonianEvolution, Observable
from pyqtorch.matrices import COMPLEX_TO_REAL_DTYPES, DEFAULT_MATRIX_DTYPE
from pyqtorch.primitives import Parametric
from pyqtorch.utils import PSR_ACCEPTANCE, GRADCHECK_sampling_ATOL


def Hamiltonian_general(n_qubits: int = 2, batch_size: int = 1) -> torch.Tensor:
    H_batch = torch.zeros(
        (2**n_qubits, 2**n_qubits, batch_size), dtype=DEFAULT_MATRIX_DTYPE
    )
    for i in range(batch_size):
        H_0 = torch.randn((2**n_qubits, 2**n_qubits), dtype=DEFAULT_MATRIX_DTYPE)
        H = (H_0 + torch.conj(H_0.transpose(0, 1))).to(DEFAULT_MATRIX_DTYPE)
        H_batch[..., i] = H
    return H_batch


def circuit_psr(n_qubits: int, different_names: bool = True) -> QuantumCircuit:
    """Helper function to make an example circuit using single gap PSR."""

    param_names = ["x", "y", "theta"]
    if not different_names:
        param_names = random_parameter_names(param_names)

    ops = [
        pyq.RX(0, param_names[0]),
        pyq.RY(1, param_names[1]),
        pyq.X(1),
        pyq.Z(0),
        pyq.Y(0),
        pyq.H(0),
        pyq.T(1),
        pyq.S(1),
        pyq.SDagger(0),
        pyq.RX(0, param_names[2]),
        pyq.RY(1, torch.pi / 2),
        pyq.CNOT(0, 1),
    ]

    circ = QuantumCircuit(n_qubits, ops)

    return circ


def circuit_gpsr(n_qubits: int, different_names: bool = True) -> QuantumCircuit:
    """Helper function to make an example circuit using multi gap GPSR."""
    param_names = ["theta" + str(i) for i in range(4)]
    if not different_names:
        param_names = random_parameter_names(param_names)

    ops = [
        pyq.Y(1),
        pyq.RX(0, param_names[0]),
        pyq.PHASE(0, param_names[1]),
        pyq.CSWAP(0, (1, 2)),
        pyq.CRX(1, 2, param_names[2]),
        pyq.X(1),
        pyq.Z(0),
        pyq.Y(0),
        pyq.H(0),
        pyq.T(1),
        pyq.S(1),
        pyq.SDagger(0),
        pyq.CPHASE(1, 2, param_names[3]),
        pyq.CNOT(0, 1),
        pyq.Toffoli((0, 1), 2),
    ]

    circ = QuantumCircuit(n_qubits, ops)

    return circ


def circuit_sequence(n_qubits: int, different_names: bool = True) -> QuantumCircuit:
    """Helper function to make an example circuit using Sequences of rotations."""
    name_angles = "theta"

    ops_rx = pyq.Sequence(
        [pyq.RX(i, param_name=name_angles + "_x_" + str(i)) for i in range(n_qubits)]
    )
    ops_rz = pyq.Sequence(
        [
            pyq.RZ(
                i,
                param_name=(
                    name_angles + "_z_" + str(i)
                    if different_names
                    else name_angles + "_z"
                ),
            )
            for i in range(n_qubits)
        ]
    )
    cnot = pyq.CNOT(1, 2)
    ops = [ops_rx, ops_rz, cnot]
    circ = QuantumCircuit(n_qubits, ops)
    return circ


def circuit_hamevo_tensor_gpsr(
    n_qubits: int,
    different_names: bool = True,
    commuting_terms: bool = False,  # unused here
) -> QuantumCircuit:
    """Helper function to make an example circuit."""

    ham = Hamiltonian_general(n_qubits)
    ham_op = pyq.HamiltonianEvolution(ham, "t", qubit_support=tuple(range(n_qubits)))
    param_names = ["theta" + str(i) for i in range(4)]
    if not different_names:
        param_names = random_parameter_names(param_names)

    ops = [
        pyq.CRX(0, 1, param_names[0]),
        pyq.X(1),
        pyq.CRY(1, 2, param_names[1]),
        ham_op,
        pyq.CRX(1, 2, param_names[2]),
        pyq.X(0),
        pyq.CRY(0, 1, param_names[3]),
        pyq.CNOT(0, 1),
    ]

    circ = QuantumCircuit(n_qubits, ops)

    return circ


def circuit_hamevo_pauligen_gpsr(
    n_qubits: int, different_names: bool = True, commuting_terms: bool = False
) -> QuantumCircuit:
    """Helper function to make an example circuit."""

    ham = random_pauli_hamiltonian(
        n_qubits,
        k_1q=n_qubits,
        k_2q=0,
        default_scale_coeffs=1.0,
        commuting_terms=commuting_terms,
    )[0]
    ham_op = pyq.HamiltonianEvolution(ham, "t", qubit_support=tuple(range(n_qubits)))

    param_names = ["theta" + str(i) for i in range(4)]
    if not different_names:
        param_names = random_parameter_names(param_names)

    ops = [
        pyq.CRX(0, 1, param_names[0]),
        pyq.X(1),
        pyq.CRY(1, 2, param_names[1]),
        ham_op,
        pyq.CRX(1, 2, param_names[2]),
        pyq.X(0),
        pyq.CRY(0, 1, param_names[3]),
        pyq.CNOT(0, 1),
    ]

    circ = QuantumCircuit(n_qubits, ops)

    return circ


@pytest.mark.parametrize(
    ["n_qubits", "batch_size", "circuit_fn"],
    [
        (3, 1, circuit_hamevo_tensor_gpsr),
        (3, 1, circuit_hamevo_pauligen_gpsr),
    ],
)
@pytest.mark.parametrize("different_names", [False, True])
@pytest.mark.parametrize("commuting_terms", [False, True])
def test_expectation_gpsr_hamevo(
    n_qubits: int,
    batch_size: int,
    circuit_fn: Callable,
    different_names: bool,
    commuting_terms: bool,
    dtype: torch.dtype = torch.complex128,
) -> None:
    torch.manual_seed(42)
    circ = circuit_fn(n_qubits, different_names, commuting_terms).to(dtype)
    obs = Observable(
        random_pauli_hamiltonian(
            n_qubits,
            k_1q=n_qubits,
            k_2q=0,
            default_scale_coeffs=1.0,
        )[0]
    ).to(dtype)
    values = {
        op.param_name: torch.rand(
            batch_size, requires_grad=True, dtype=COMPLEX_TO_REAL_DTYPES[dtype]
        )
        for op in circ.flatten()
        if isinstance(op, Parametric) and isinstance(op.param_name, str)
    }
    values.update(
        {
            op.time: torch.rand(
                batch_size, requires_grad=True, dtype=COMPLEX_TO_REAL_DTYPES[dtype]
            )
            for op in circ.operations
            if isinstance(op, HamiltonianEvolution) and isinstance(op.time, str)
        }
    )
    state = pyq.random_state(n_qubits, dtype=dtype)

    # Apply adjoint
    exp_ad = expectation(circ, state, values, obs, DiffMode.AD)
    grad_ad = torch.autograd.grad(
        exp_ad, tuple(values.values()), torch.ones_like(exp_ad), create_graph=True
    )

    # Apply PSR
    exp_gpsr = expectation(circ, state, values, obs, DiffMode.GPSR)
    grad_gpsr = torch.autograd.grad(
        exp_gpsr, tuple(values.values()), torch.ones_like(exp_gpsr), create_graph=True
    )

    # first order checks
    for i in range(len(grad_ad)):
        assert torch.allclose(grad_ad[i], grad_gpsr[i], atol=PSR_ACCEPTANCE)

    # second order checks
    for i in range(len(grad_ad)):
        gradgrad_ad = torch.autograd.grad(
            grad_ad[i],
            tuple(values.values()),
            torch.ones_like(grad_ad[i]),
            create_graph=True,
        )

        gradgrad_gpsr = torch.autograd.grad(
            grad_gpsr[i],
            tuple(values.values()),
            torch.ones_like(grad_gpsr[i]),
            create_graph=True,
        )

        assert len(gradgrad_ad) == len(gradgrad_gpsr)

        # check second order gradients
        for j in range(len(gradgrad_ad)):
            assert torch.allclose(gradgrad_ad[j], gradgrad_gpsr[j], atol=1.0e-2)


@pytest.mark.parametrize(
    ["n_qubits", "batch_size", "circuit_fn"],
    [
        (2, 1, circuit_psr),
        (5, 10, circuit_psr),
        (3, 1, circuit_gpsr),
        (5, 10, circuit_gpsr),
        (3, 1, circuit_sequence),
        (5, 10, circuit_sequence),
    ],
)
@pytest.mark.parametrize("dtype", [torch.complex64, torch.complex128])
@pytest.mark.parametrize("different_names", [False, True])
def test_expectation_gpsr(
    n_qubits: int,
    batch_size: int,
    circuit_fn: Callable,
    dtype: torch.dtype,
    different_names: bool,
) -> None:
    torch.manual_seed(42)
    circ = circuit_fn(n_qubits, different_names).to(dtype)
    obs = Observable(
        random_pauli_hamiltonian(
            n_qubits,
            k_1q=n_qubits,
            k_2q=0,
            default_scale_coeffs=1.0,
        )[0]
    ).to(dtype)
    values = {
        op.param_name: torch.rand(
            batch_size, requires_grad=True, dtype=COMPLEX_TO_REAL_DTYPES[dtype]
        )
        for op in circ.flatten()
        if isinstance(op, Parametric) and isinstance(op.param_name, str)
    }
    state = pyq.random_state(n_qubits, dtype=dtype)

    # Apply adjoint
    exp_ad = expectation(circ, state, values, obs, DiffMode.AD)
    grad_ad = torch.autograd.grad(
        exp_ad, tuple(values.values()), torch.ones_like(exp_ad), create_graph=True
    )

    # Apply PSR
    exp_gpsr = expectation(circ, state, values, obs, DiffMode.GPSR)
    grad_gpsr = torch.autograd.grad(
        exp_gpsr, tuple(values.values()), torch.ones_like(exp_gpsr), create_graph=True
    )

    exp_gpsr_sampled = expectation(
        circ,
        state,
        values,
        obs,
        DiffMode.GPSR,
        n_shots=100000,
    )
    grad_gpsr_sampled = torch.autograd.grad(
        exp_gpsr_sampled,
        tuple(values.values()),
        torch.ones_like(exp_gpsr_sampled),
        create_graph=True,
    )
    assert torch.allclose(exp_gpsr, exp_gpsr_sampled, atol=1e-01)

    # first order checks
    for i in range(len(grad_ad)):
        assert torch.allclose(grad_ad[i], grad_gpsr[i], atol=PSR_ACCEPTANCE)
        assert torch.allclose(
            grad_gpsr[i], grad_gpsr_sampled[i], atol=GRADCHECK_sampling_ATOL
        )

    # second order checks
    for i in range(len(grad_ad)):
        gradgrad_ad = torch.autograd.grad(
            grad_ad[i],
            tuple(values.values()),
            torch.ones_like(grad_ad[i]),
            create_graph=True,
        )

        gradgrad_gpsr = torch.autograd.grad(
            grad_gpsr[i],
            tuple(values.values()),
            torch.ones_like(grad_gpsr[i]),
            create_graph=True,
        )

        assert len(gradgrad_ad) == len(gradgrad_gpsr)

        # check second order gradients
        for j in range(len(gradgrad_ad)):
            assert torch.allclose(gradgrad_ad[j], gradgrad_gpsr[j], atol=PSR_ACCEPTANCE)


@pytest.mark.parametrize("gate_type", ["scale", "hamevo", "same", ""])
@pytest.mark.parametrize("sequence_circuit", [True, False])
def test_compatibility_gpsr(gate_type: str, sequence_circuit: bool) -> None:

    pname = "theta_0"
    if gate_type == "scale":
        seq_gate = pyq.Sequence([pyq.X(0)])
        scale = pyq.Scale(seq_gate, pname)
        ops = [scale]
    elif gate_type == "hamevo":
        symbol = pname
        t_evo = torch.tensor([torch.pi / 4], dtype=DEFAULT_MATRIX_DTYPE)
        hamevo = pyq.HamiltonianEvolution(symbol, t_evo, (0,))
        ops = [hamevo]
    elif gate_type == "same":
        ops = [pyq.RY(0, pname), pyq.RZ(0, pname)]
    else:
        # check that CNOT is not tested on spectral gap call
        ops = [pyq.RY(0, pname), pyq.CNOT(0, 1)]

    if sequence_circuit:
        circ = pyq.QuantumCircuit(2, pyq.Sequence(ops))
    else:
        circ = pyq.QuantumCircuit(2, ops)
    obs = pyq.Observable([pyq.Z(0)])
    state = pyq.zero_state(2)

    if gate_type == "hamevo":
        H = pyq.X(0).tensor()
        H.requires_grad = True
        values = {"theta_0": H}
    else:
        param_value = torch.pi / 2
        values = {"theta_0": torch.tensor([param_value], requires_grad=True)}

    if gate_type not in ("", "same"):
        with pytest.raises(ValueError):
            exp_gpsr = expectation(circ, state, values, obs, DiffMode.GPSR)
    else:
        exp_gpsr = expectation(circ, state, values, obs, DiffMode.GPSR)

        grad_gpsr = torch.autograd.grad(
            exp_gpsr, tuple(values.values()), torch.ones_like(exp_gpsr)
        )
        assert len(grad_gpsr) == 1
